
__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import radical.utils.config as ruc


# ------------------------------------------------------------------------------
def test_config () :
    """ Test if configuration location is picked up from env """

    tmp = '/tmp/radical.util.cfg.%s' % os.getpid ()
    os.environ ['RADICAL_UTILS_CONFIG'] = tmp

    with  open (tmp, 'w') as tmp_file :
        tmp_file.write ("[test]\n")
        tmp_file.write ("key=val\n")


    # set the configuration options for this object
    configurable = ruc.Configurable        ('radical.utils')
    config       = configurable.get_config ('test')

    assert ('val' == config['key'].get_value ())

    os.remove (tmp)

# ------------------------------------------------------------------------------
#
# run tests if called directly
#
if __name__ == "__main__":

    test_config ()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import imp
import sys
import glob

import singleton

from .logger import get_logger


# ------------------------------------------------------------------------------
#
class _PluginRegistry (dict) :
    """
    The plugin registry helper class avoids that plugins are loaded twice.
    """

    __metaclass__ = singleton.Singleton


    # --------------------------------------------------------------------------
    #
    def __init__ (self) :

        self._registry = dict ()


    # --------------------------------------------------------------------------
    #
    def register (self, namespace, plugins) :

        if  not namespace in self._registry :
            self._registry[namespace] = plugins


    # --------------------------------------------------------------------------
    #
    def retrieve (self, namespace) :

        if  namespace in self._registry :
            return self._registry[namespace]

        return None


# ------------------------------------------------------------------------------
#
class PluginManager (object) :
    """ 
    The RADICAL plugin management and loading utility.

    The plugin manager allows to manage plugins of a specific types.  For those
    types, the manager can search for installed plugins, list and describe
    plugins found, load plugins, and instantiate the plugin for further use.

    Example::

        # try to load the 'echo' plugin from the 'radical' namespace
        plugin_type = 'echo'

        pm = radical.utils.PluginManager ('radical')

        for plugin_name in pm.list (plugin_type) :
            print plugin_name
            print pm.describe (plugin_type, plugin_name)

        default_plugin = pm.load ('echo', 'default')

        default_plugin.init_plugin ("world")
        default_plugin.run ()  # prints "hello default world"


    The plugins are expected to follow a specific naming and coding schema to be
    recognized by the plugin manager.  The naming schema is:

        [namespace].plugins.[ptype].plugin_[ptype]_[pname].py

    i.e. for the example above: `radical.plugins.echo.plugin_echo_default.py`

    The plugin code consists of two parts:  a plugin description, and a plugin
    class.  The description is a module level dictionary named
    `PLUGIN_DESCRIPTION`, the plugin class must be named `PLUGIN_CLASS`, and
    must have a class constructor `__init__(*args, **kwargs)` to create plugin
    instances for further use.

    At this point, we leave the definition of the exact plugin signatures open,
    but expect that to be more strictly defined per plugin type in the future.

    Note that the PluginManager construction is, at this point, not considered
    thread safe.
    """


    #---------------------------------------------------------------------------
    # 
    def __init__ (self, namespace) :
        """
        namespace: name of module (plugins are expected in namespace/plugins/)
        """

        # import here to avoid circular imports
        import radical.utils.logger as logger

        self._namespace = namespace
        self._logger    = get_logger('radical.utils')
        self._registry  = _PluginRegistry ()  # singleton
        self._plugins   = self._registry.retrieve (self._namespace)

        # load adaptors if registry didn't have any registered, yet
        if  not self._plugins :
            self._load_plugins ()
            self._registry.register (self._namespace, self._plugins)


    #---------------------------------------------------------------------------
    # 
    def _load_plugins (self) :
        """ 
        Load all plugins for the given namespace.  Previously loaded plugins
        are overloaded.
        """

        # start wish a fresh plugin registry
        self._plugins = dict () 

        self._logger.info ('loading plugins for namespace %s' % self._namespace)

        # avoid to load plugins twice in case of redundant sys paths
        seen = list() 

        # search for plugins in all system module paths
        for spath in sys.path :

            # we only load plugins installed under the namespace hierarchy
            npath = self._namespace.replace ('.', '/')
            ppath = "%s/%s/plugins/"  %  (spath, npath)
            pglob = "*/plugin_*.py"  

            # make sure the 'plugins' dir exists
            if  not os.path.isdir (ppath) :
                continue

            # we assume that all python sources in that location are
            # suitable plugins
            pfiles = glob.glob (ppath + pglob)

            if  not pfiles :
                continue

            for pfile in pfiles :

                # from the full plugin file name, derive a short name for more
                # useful logging, duplication checks etc. -- simply remove the
                # namespace path portion...
                if  pfile.startswith (spath) :
                    pshort = pfile[len(spath):]
                else :
                    pshort = pfile

              # print "pfile : %s" % pfile
              # print "npath : %s" % npath
              # print "spath : %s" % spath
              # print "pshort: %s" % pshort

                # check for duplication
                if  pshort in seen :
                    continue
                else :
                    seen.append (pshort)

                # modname for 'load_source' needs to be unique, otherwise global
                # vars in the plugin file (such as, aehm, PLUGIN_DESCRIPTION)
                # will be overwritten by the next plugin load.
                pmodname = "%s.plugins.%s" % (self._namespace, 
                                              os.path.basename(os.path.dirname(pfile)))
                modname  = "%s.%s"         % (pmodname,
                                              os.path.splitext(os.path.basename(pfile))[0])

                try :
                    # load and register the plugin

                    # modname is unique and correct -- but load_source raises
                    # a RuntimeWarning, because, apparently, the plugin's parent
                    # module cannot be found.  In fact, the parent is a proper
                    # python module, and it loads fine via 'import' -- but it is
                    # is not imported before, load_source doesn't like that.
                    # Well, the parent module actually SHOULD NOT be imported --
                    # but we do it here anyways to silence the warning.  Thanks
                    # python...
                    __import__ (pmodname)

                    # now load the plugin proper
                    plugin = imp.load_source (modname, pfile)

                    # get plugin details from description
                    ptype  = plugin.PLUGIN_DESCRIPTION.get ('type',        None)
                    pname  = plugin.PLUGIN_DESCRIPTION.get ('name',        None)
                    pvers  = plugin.PLUGIN_DESCRIPTION.get ('version',     None)
                    pdescr = plugin.PLUGIN_DESCRIPTION.get ('description', None)

                    # make sure details are complete
                    if  not ptype  : 
                        self._logger.error ('no plugin type in %s' % pshort)
                        continue

                    if  not pname  : 
                        self._logger.error ('no plugin name in %s' % pshort)
                        continue

                    if  not pvers  : 
                        self._logger.error ('no plugin version in %s' % pshort)
                        continue

                    if  not pdescr : 
                        self._logger.error ('no plugin description in %s' % pshort)
                        continue

                    # now put the plugin and plugin info into the plugin
                    # registry.  Duh!
                    if  not ptype in self._plugins :
                        self._plugins[ptype] = {}

                    if  pname in self._plugins[ptype] :
                        self._logger.warn ('overloading plugin %s' % pshort)

                    self._plugins[ptype][pname] = {
                        'class'       : plugin.PLUGIN_CLASS,
                        'type'        : ptype, 
                        'name'        : pname, 
                        'version'     : pvers, 
                        'description' : pdescr,
                        'instance'    : None
                    }

                    self._logger.debug ('loading plugin %s' % pfile)
                    self._logger.info  ('loading plugin %s' % pshort)

                except Exception as e :
                    self._logger.error ('loading plugin %s failed: %s' % (pshort, e))


    #---------------------------------------------------------------------------
    # 
    def list_types (self) :
        """
        return a list of loaded plugin types
        """
        return self._plugins.keys ()


    #---------------------------------------------------------------------------
    # 
    def list (self, ptype) :
        """
        return a list of loaded plugins for a given plugin type
        """
        if  not ptype in self._plugins :
            self._logger.debug (self.dump_str())
            raise LookupError ("No such plugin type %s in %s" \
                    % (ptype, self._plugins.keys()))

        return self._plugins[ptype].keys ()


    #---------------------------------------------------------------------------
    # 
    def describe (self, ptype, pname) :
        """
        return a plugin details for a given plugin type / name pair
        """
        if  not ptype in self._plugins :
            self._logger.debug (self.dump_str())
            raise LookupError ("No such plugin type %s in %s" \
                    % (ptype, self._plugins.keys()))

        if  not pname in self._plugins[ptype] :
            self._logger.debug (self.dump_str())
            raise LookupError ("No such plugin name %s (type: %s) in %s" \
                    % (pname, ptype, self._plugins[ptype].keys()))

        return self._plugins[ptype][pname]


    #---------------------------------------------------------------------------
    # 
    def load (self, ptype, pname) :
        """
        check if a plugin with given type and name was loaded, if so, instantiate its
        plugin class and return it.
        """

        if  not ptype in self._plugins :
            self._logger.debug (self.dump_str())
            raise LookupError ("No such plugin type %s in %s" \
                    % (ptype, self._plugins.keys()))

        if  not pname in self._plugins[ptype] :
            self._logger.debug (self.dump_str())
            raise LookupError ("No such plugin name %s (type: %s) in %s" \
                    % (pname, ptype, self._plugins[ptype].keys()))

        # create new plugin instance
        return self._plugins[ptype][pname]['class']()


    #---------------------------------------------------------------------------
    # 
    def dump (self) :

        import pprint
        pprint.pprint (self._plugins)


    #---------------------------------------------------------------------------
    # 
    def dump_str (self) :

        import pprint
        return "\n%s" % pprint.pformat (self._plugins)


# ------------------------------------------------------------------------------





""" 
Provide an abstraction for distributed queue network.  

The network nodes are instances of queue_network::Node.  A node can exist either
as thread, as a local process, or as a remote process, the latter spawned via
SAGA.

The communication between nodes is always via queues, were
groups of nodes feed into the same queues, and/or are fed from the same queues.
A node can be fed from multiple queues, where queues are prioritized according
to some policy.  Depending if two communicating nodes are co-threads,
co-processes, or remote processes, different queue implementations (with the
same semantic) are used.

A network is created according to a specific state model.  That model describes
the state of entities passing through the queue network.  Each active state is
enacted by a network node, and each pending state is represented by the entity
being passed through a queue.  A queue connection thus exists for every valid
state transition.

Nodes also support the observer patter, ie. they can notify observers about
enacted state transitions. 
"""

# ------------------------------------------------------------------------------
#
# http://stackoverflow.com/questions/9539052/python-dynamically-changing-base-classes-at-runtime-how-to
#
# Depending on agent architecture (which is specific to the resource type it
# runs on) can switch between different component types: using threaded (when
# running on the same node), multiprocessing (also for running on the same node,
# but avoiding python's threading problems, for the prices of slower queues),
# and remote processes (for running components on different nodes, using zeromq
# queues for communication).
#
# We do some trickery to keep the actual components independent from the actual
# schema:
#
#   - we wrap the different queue types into a rpu.Queue object
#   - we change the base class of the component dynamically to the respective type
#
# This requires components to adhere to the following restrictions:
#
#   - *only* communicate over queues -- no shared data with other components or
#     component instances.  Note that this also holds for example for the
#     scheduler!
#   - no shared data between the component class and it's run() method.  That
#     includes no sharing of queues.
#   - components inherit from base_component, and the constructor needs to
#     register all required component-internal and -external queues with that
#     base class -- the run() method can then transparently retrieve them from
#     there.
#


# ------------------------------------------------------------------------------
#
class network (object):

    # --------------------------------------------------------------------------
    #
    def __init__ (self, state_model={}, network_description={}):
        """
        span the network according to state model and network description
        """

        self.state_model = state_model


    # --------------------------------------------------------------------------
    #
    def feed (self, entities):
        """
        feed entities into the network
        """

        if not isinstance(entities, list):
            entities = [entities]

        for entity in entities:
            assert (entity.state == self.state_model['initial'])

        self.feeder_queue.push (entities)


    # --------------------------------------------------------------------------
    #
    def subscribe (self, states, callback):
        """
        subscribe for callback notifications for state transitions *into* the
        specified states.  If callbacks return 'False' or 'None', they are
        unregistered.
        """

        if not isinstance(states, list):
            states = [states]







#!/usr/bin/env python

import os
import sys
import time
import socket
import threading as mt

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_poll():

    sp = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)
    parent = os.getpid()
    child  = os.fork()

    if not child:

      # print 'child : %s' % os.getpid()
    
        sp[0].close()
        sp[1].send('foo')
        time.sleep(1)
        sp[1].send('bar')
        time.sleep(2)
        sp[1].close()
      # print 'child : done'
        sys.exit()
    
    else:

        sp[1].close()

        poller = ru.Poller()
        poller.register(sp[0], ru.POLLERR | ru.POLLIN | ru.POLLHUP)

      # print 'Parent: %s' % parent
      # print 'Child : %s' % child

        abort = False
        msgs  = list()
        while not abort:

            events = poller.poll(1.0)

            for fd,event in events:
                
                if event & ru.POLLIN:
                    msg = sp[0].recv(200)
                    assert(msg in ['foo', 'bar', ''])
                    msgs.append(msg)

                if  event & ru.POLLERR or event & ru.POLLHUP:
                    abort = True
                    assert(msgs[0] == 'foo')
                    assert(msgs[1] == 'bar')
                    assert(msgs[2] == '')
                    break

        os.waitpid(child, 0)
      # print 'child : %s gone' % child


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_poll()


# ------------------------------------------------------------------------------






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_ids () :
    """
    Test ID generation
    """

    id_a_1 = ru.generate_id ('a')
    id_a_2 = ru.generate_id ('a')
    id_a_3 = ru.generate_id ('a', mode=ru.ID_SIMPLE)
    id_b_1 = ru.generate_id ('b', mode=ru.ID_SIMPLE)
    id_b_2 = ru.generate_id ('b', mode=ru.ID_SIMPLE)
    id_c_1 = ru.generate_id ('c', mode=ru.ID_UNIQUE)
    id_c_2 = ru.generate_id ('c', mode=ru.ID_UNIQUE)
    id_d_1 = ru.generate_id ('d.%(day_counter)03d', mode=ru.ID_CUSTOM)
    id_d_2 = ru.generate_id ('d.%(day_counter)03d', mode=ru.ID_CUSTOM)
    id_e_1 = ru.generate_id ('e', mode=ru.ID_UUID)
    id_e_2 = ru.generate_id ('e', mode=ru.ID_UUID)

    assert (id_a_1 == 'a.0000'       ), "'%s' == 'a.0000'"       % (id_a_1)
    assert (id_a_2 == 'a.0001'       ), "'%s' == 'a.0001'"       % (id_a_2)
    assert (id_a_3 == 'a.0002'       ), "'%s' == 'a.0002'"       % (id_a_3)
    assert (id_b_1 == 'b.0000'       ), "'%s' == 'b.0000'"       % (id_b_1)
    assert (id_b_2 == 'b.0001'       ), "'%s' == 'b.0001'"       % (id_b_2)
    assert (id_c_1.startswith  ('c.')), "'%s'.startswith ('c.')" % (id_c_1)
    assert (id_c_2.startswith  ('c.')), "'%s'.startswith ('c.')" % (id_c_2)
    assert (id_c_1.endswith ('.0000')), "'%s'.endswith ('0000')" % (id_c_1)
    assert (id_c_2.endswith ('.0001')), "'%s'.endswith ('0001')" % (id_c_2)
    assert (id_d_1.startswith  ('d.')), "'%s'.startswith ('d.')" % (id_d_1)
    assert (id_d_2.startswith  ('d.')), "'%s'.startswith ('d.')" % (id_d_2)
    assert (id_e_1.startswith  ('e.')), "'%s'.startswith ('e.')" % (id_e_1)
    assert (id_e_2.startswith  ('e.')), "'%s'.startswith ('e.')" % (id_e_2)

    assert (id_a_1 != id_a_2         ), "'%s' != '%s'"   % (id_a_1, id_a_2)
    assert (id_a_1 != id_a_3         ), "'%s' != '%s'"   % (id_a_1, id_a_3)
    assert (id_a_2 != id_a_3         ), "'%s' != '%s'"   % (id_a_2, id_a_3)
    assert (id_b_1 != id_b_2         ), "'%s' != '%s'"   % (id_b_1, id_b_2)
    assert (id_c_1 != id_c_2         ), "'%s' != '%s'"   % (id_c_1, id_c_2)
    assert (id_d_1 != id_d_2         ), "'%s' != '%s'"   % (id_d_1, id_d_2)
    assert (id_e_1 != id_e_2         ), "'%s' != '%s'"   % (id_e_1, id_e_2)

    assert (isinstance (id_a_1, basestring)), "isinstance('%s', basestring)" % (id_a_1)
    assert (isinstance (id_a_2, basestring)), "isinstance('%s', basestring)" % (id_a_2)
    assert (isinstance (id_a_3, basestring)), "isinstance('%s', basestring)" % (id_a_3)
    assert (isinstance (id_b_1, basestring)), "isinstance('%s', basestring)" % (id_b_1)
    assert (isinstance (id_b_2, basestring)), "isinstance('%s', basestring)" % (id_b_2)
    assert (isinstance (id_c_1, basestring)), "isinstance('%s', basestring)" % (id_c_1)
    assert (isinstance (id_c_2, basestring)), "isinstance('%s', basestring)" % (id_c_2)
    assert (isinstance (id_c_1, basestring)), "isinstance('%s', basestring)" % (id_c_1)
    assert (isinstance (id_c_2, basestring)), "isinstance('%s', basestring)" % (id_c_2)
    assert (isinstance (id_d_1, basestring)), "isinstance('%s', basestring)" % (id_d_1)
    assert (isinstance (id_e_1, basestring)), "isinstance('%s', basestring)" % (id_e_1)

    try                   : id_x = ru.generate_id (None)
    except TypeError      : pass
    except Exception as e : assert (False), "TypeError  != %s" % type(e)

    try                   : id_x = ru.generate_id (1)
    except TypeError      : pass
    except Exception as e : assert (False), "TypeError  != %s" % type(e)

    try                   : id_x = ru.generate_id ('a.', mode='RANDOM')
    except ValueError     : pass
    except Exception as e : assert (False), "ValueError != %s" % type(e)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_ids ()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


from node    import Node
from queue   import Queue
from network import Network


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_dict_mixin () :
    """
    Test dict mixin
    """

    # --------------------------------------------------------------------------
    @ru.Lockable
    class Test (ru.DictMixin) :

        # ----------------------
        def __init__ (self) :

            self._d     = dict ()
            self['val'] = 1

        def __getitem__(self, key):
            return self._d.__getitem__ (key)

        def __setitem__(self, key, val):
            return self._d.__setitem__ (key, val)

        def __delitem__(self, key):
            return self._d.__delitem__ (key)

        def keys(self):
            return self._d.keys ()


    # --------------------------------------------------------------------------
    t = Test ()

    assert (t['val']        == 1       )
    assert (t.keys()        == ['val'] )

    assert ('val'       in t)
    assert ('test1' not in t)
    assert ('test2' not in t)

    t['test1'] =  'test'
    t['test2'] = ['test']

    assert ('val'       in t)
    assert ('test1'     in t)
    assert ('test2'     in t)

    assert (t['val']        == 1       )
    assert (t['test1']      == 'test'  )
    assert (t['test2']      == ['test'])
    assert (t.keys().sort() == ['val', 'test1', 'test2'].sort()), "%s" % str(t.keys())

    del t['test1']

    assert (t['val']        == 1       )
    assert (t['test2']      == ['test'])
    assert (t.keys().sort() == ['val', 'test2'].sort()), "%s" % str(t.keys())

    assert ('val'       in t)
    assert ('test1' not in t)
    assert ('test2'     in t)

# ------------------------------------------------------------------------------
#
def test_dict_merge () :

    dict_1 = {'key_shared' : 'val_shared_1', 
              'key_orig_1' : 'val_orig_1'}
    dict_2 = {'key_shared' : 'val_shared_2', 
              'key_orig_2' : 'val_orig_2'}

    try :
        ru.dict_merge (dict_1, dict_2)
        assert (False), 'expected ValueError exception'
    except ValueError :
        pass
    except Exception as e :
        assert (False), 'expected ValueError exception, not %s' % e


    ru.dict_merge (dict_1, dict_2, policy='preserve')

    assert (dict_1.keys()        == ['key_orig_1', 'key_orig_2', 'key_shared'])
    assert (dict_1['key_shared'] == 'val_shared_1')
    assert (dict_1['key_orig_1'] == 'val_orig_1')
    assert (dict_1['key_orig_2'] == 'val_orig_2')


    ru.dict_merge (dict_1, dict_2, policy='overwrite')

    assert (dict_1.keys()        == ['key_orig_1', 'key_orig_2', 'key_shared'])
    assert (dict_1['key_shared'] == 'val_shared_2')
    assert (dict_1['key_orig_1'] == 'val_orig_1')
    assert (dict_1['key_orig_2'] == 'val_orig_2')


# ------------------------------------------------------------------------------
#
def test_dict_stringexpand () :

    target = {'workdir'  : '%(home)s/work/',
              'resource' : '%(resource)s'}
    source = {'user'     : 'peer_gynt',
              'protocol' : 'ssh',
              'host'     : 'localhost',
              'home'     : '/home/%(user)s', 
              'resource' : '%(protocol)s://%(host)s/'}

    ru.dict_stringexpand (target, source)

    assert (target.keys()      == ['workdir', 'resource'])
    assert (target['workdir']  == '/home/peer_gynt/work/')
    assert (target['resource'] == 'ssh://localhost/')

# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_dict_mixin ()
    test_dict_merge ()
    test_dict_stringexpand ()

# ------------------------------------------------------------------------------





import os
import regex


# ------------------------------------------------------------------------------
#
def collapse_ranges (ranges):
    """
    Given be a set of ranges (as a set of pairs of floats [start, end] with
    'start <= end'.  This algorithm will then collapse that set into the
    smallest possible set of ranges which cover the same, but not more nor
    less, of the domain (floats).

    We first sort the ranges by their starting point.  We then start with the
    range with the smallest starting point [start_1, end_1], and compare to
    the next following range [start_2, end_2], where we now know that
    start_1 <= start_2.  We have now two cases:

    a) when start_2 <= end_1, then the ranges overlap, and we collapse them into
    range_1: range_1 = [start_1, max[end_1, end_2]

    b) when start_2 > end_1, then ranges don't overlap.  Importantly, none of
    the other later ranges can ever overlap range_1, because there start points
    are even larger.  So we move range_1 to the set of final ranges, and restart
    the algorithm with range_2 being the smallest one.

    Termination condition is if only one range is left -- it is also moved to
    the list of final ranges then, and that list is returned.
    """

    # FIXME: does tuple and set conversion really add anything?

    # Ranges must be unique: we do not count timings when they start and end at
    # exactly the same time. By using a set, we do not repeat ranges.
    # we convert to a list before return.
    final = set()

    # return empty list if given an empty list
    if not ranges:
        return []

    START = 0
    END   = 1

    # sort ranges into a copy list, by their start time
    _ranges = sorted(ranges, key=lambda x: x[START])

    # sat 'base' to the earliest range (smallest starting point)
    base = _ranges[0]

    for _range in _ranges[1:]:

        # if range is empty, skip it
        if _range[START] == _range[END]:
            continue

        if _range[START] <= base[END]:
            # ranges overlap -- extend the base
            base[END] = max(base[END], _range[END])

        else:
            # ranges don't overlap -- move base to final, and current _range
            # becomes the new base
            final.add(tuple(base))
            base = _range

    # termination: push last base to final
    final.add(tuple(base))

    # Return final as list of list in case a mutable type is needed.
    return [list(b) for b in final]


# ------------------------------------------------------------------------------
#
def range_concurrency(ranges):
    '''
    given a set of *un*collapsed ranges, return a series which describes the
    range-concurrency at any point.

    Example:
      Ranges: 
        [----]  [---------]             [--------]
          [--------]         [-]      [------------]      [--]

      Concurrency:
        1 2  1  2  1      0  1 0      1 2        1 0      1  0

    Returned is a sorted list of tuples where the first entry defines at what
    range value the concurrency changed, and the second value defines to what
    the concurrency count changed at that point.

    You could consider the ranges to be of type `[time_start, time_end]`, and
    the return would be a list of `[timestamp, concurrency]`, if that helps --
    but the algorithm does not make any assumption on the data type, really,
    only that the values can be sorted.
    '''

    if not ranges:
        return list()

    # we want to sort all range boundaries by value, but also want to remember
    # if they were start or end
    START = 0
    END   = 1
    vals  = list()
    for idx in range(len(ranges)):
        r = ranges[idx]
        vals.append([r[0], START])
        vals.append([r[1], END  ])

    # go through the sorted list of times, and increase concurrency on `START`,
    # decrease it on `END`.  Make sure we add up all entries for the same value
    # at once.

    # first entry
    assert(vals[0][1] == START), 'inconsistent ranges'
    last_val    = vals[0][0]
    ret         = list()
    concurrency = 1
    for val,kind in sorted(vals):
        if val == last_val:
            last_val = val
            if kind == START: concurrency += 1
            else            : concurrency -= 1
        else:
            ret.append([last_val, concurrency])

    assert(concurrency == 0), 'inconsistent range structure?'
    ret.append([last_val, concurrency])

    return ret


# ------------------------------------------------------------------------------
#
def partition(space, nparts):
    '''
    create balanced partitions from an iterable space.  This method preserves
    contiguous order.

    kudos:
    http://code.activestate.com/recipes/425397-split-a-list-into-roughly-equal-sized-pieces/
    '''

    n   = len(space)
    b   = 0
    ret = list()

    for k in range(nparts):

        q, r = divmod(n-k, nparts)
        a, b = b, b + q + (r!=0)

        ret.append(space[a:b])

    return ret


# ------------------------------------------------------------------------------
#
def in_range(value, ranges):
    """
    checks if a float value `value` is in any of the given `ranges`, which are
    assumed to be a list of tuples (or a single tuple) of start end end points
    (floats).
    Returns `True` or `False`.
    """

    # is there anythin to check?
    if not ranges or not len(ranges):
        return False

    if not isinstance(ranges[0], list):
        ranges = [ranges]

    for r in ranges:
        if value >= r[0] and value <= r[1]:
            return True

    return False


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test = [ [ 0, 10],
             [20, 30],
             [40, 50],
             [60, 70],
             [80, 90],
             [ 5, 15],
             [35, 55] ]

    import pprint
    pprint.pprint (test)
    pprint.pprint (collapse_ranges (test))

    space = range(75)
    parts = partition(space, 8)
    for part in parts:
        print "%3d: %s" % (len(part), part)


# ------------------------------------------------------------------------------




#!/usr/bin/env python


import os
import sys
import pprint
import radical.utils as ru


_DEFAULT_MODE  = 'list'
_DEFAULT_DBURL = 'mongodb://localhost:27017/'
_DEFAULT_DBURL = 'mongodb://ec2-184-72-89-141.compute-1.amazonaws.com:27017/'

if  'RADICAL_PILOT_DBURL' in os.environ :
    _DEFAULT_DBURL = os.environ['RADICAL_PILOT_DBURL']


# ------------------------------------------------------------------------------
#
def usage (msg=None, noexit=False) :

    if  msg :
        print "\n\t%s\n" % msg

    print """

      usage   : %s -m mode [-d url]
      example : %s mongodb://localhost/synapse_profiles/profiles/

      The URL is interpreted as:
          [schema]://[host]:[port]/[database]/[collection]/[document_id]

      modes are:

        help:   show this message
        tree:   show a tree of the hierarchy, but only  document IDs, no content
        dump:   show a tree of the hierarchy, including document contents
        list:   list entries in the subtree, but do not traverse
        remove: remove the specified subtree

      The default command is 'tree'.  
      The default MongoDB is """ + "'%s'\n\n" % _DEFAULT_DBURL

    if  msg :
        sys.exit (1)

    if  not noexit :
        sys.exit (0)


# ------------------------------------------------------------------------------
#
def dump (url, mode) :
    """
    Connect to mongodb at the given location, and traverse the data bases
    """

    mongo, db, dbname, cname, pname = ru.mongodb_connect (url, _DEFAULT_DBURL)

    print dbname
 
    if  dbname : dbnames = [dbname]
    else       : dbnames = mongo.database_names ()

    for name in dbnames :

        if  mode == 'list' and not dbname :
            print " +-- db   %s" % name

        elif  mode == 'remove' :
            
            if (not dbname) or (name == dbname) :
                try :
                    mongo.drop_database (name)
                    print "  removed database %s" % name
                except :
                    pass # ignore system tables

        else :
            handle_db (mongo, mode, name, cname, pname)

    mongo.disconnect ()


# ------------------------------------------------------------------------------
def handle_db (mongo, mode, dbname, cname, pname) :
    """
    For the given db, traverse collections
    """

    database = mongo[dbname]
    print " +-- db   %s" % dbname


    if  cname : cnames = [cname]
    else      : cnames = database.collection_names ()

    for name in cnames :

        if  mode == 'list' and not cname :
            print " | +-- coll %s" % name

        elif  mode == 'remove' and not pname :
            try :
                database.drop_collection (name)
                print "  removed collection %s" % name
            except :
                pass # ignore errors

        else :
            handle_coll (database, mode, name, pname)



# ------------------------------------------------------------------------------
def handle_coll (database, mode, cname, pname) :
    """
    For a given collection, traverse all documents
    """

    if 'indexes' in cname :
        return

    collection = database[cname]
    print " | +-- coll %s" % cname

    docs = collection.find ()

    for doc in docs :

        name = doc['_id']

        if  mode == 'list' and not pname :
            print " | | +-- doc  %s" % name

        elif  mode == 'remove' :
            if (not pname) or (str(name)==str(pname)) :
                try :
                    collection.remove (name)
                    print "  removed document %s" % name
                except Exception as e:
                    pass # ignore errors

        else :
            if (not pname) or (str(name)==str(pname)) :
                handle_doc (collection, mode, doc)


# ------------------------------------------------------------------------------
def handle_doc (collection, mode, doc) :
    """
    And, surprise, for a given document, show it according to 'mode'
    """

    name = doc['_id']

    if  mode == 'list' :

        for key in doc :
            print " | | | +-- %s" % (key)

    elif  mode == 'tree' :
        print " | | +-- doc  %s" % (name)
        for key in doc :
            print " | | | +-- %s" % (key)

    elif  mode == 'dump' :
        print " | | +-- doc  %s" % (name)
        for key in doc :
            txt_in  = pprint.pformat (doc[key])
            txt_out = ""
            lnum    = 1
            for line in txt_in.split ('\n') :
                if  lnum != 1 :
                    txt_out += ' | | | |                '
                txt_out += line
                txt_out += '\n'
                lnum    += 1

            print " | | | +-- %-10s : %s" % (key, txt_out[:-1]) # remove last \n


# ------------------------------------------------------------------------------
#
if __name__ == '__main__' :

    import optparse
    parser = optparse.OptionParser (add_help_option=False)

    parser.add_option('-m', '--mode',    dest='mode')
    parser.add_option('-d', '--dburl',   dest='dburl')
    parser.add_option('-h', '--help',    dest='help', action="store_true")

    options, args = parser.parse_args ()

    if  args         : usage ("Too many arguments (%s)" % args) 
    if  options.help : usage ()

    mode    = options.mode 
    dburl   = options.dburl

    if not mode  : mode  = _DEFAULT_MODE
    if not dburl : dburl = _DEFAULT_DBURL

    print "modes   : %s" % mode
    print "db url  : %s" % dburl

    for m in mode.split (',') :

        if  m not in ['list', 'dump', 'tree', 'remove', 'help'] : 
            usage ("Unsupported mode '%s'" % m)

        elif m == 'tree'   : dump  (dburl, m) 
        elif m == 'dump'   : dump  (dburl, m) 
        elif m == 'list'   : dump  (dburl, m) 
        elif m == 'remove' : dump  (dburl, m) 
        elif m == 'help'   : usage (noexit=True)
        else               : usage ("unknown mode '%s'" % mode)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import time
import threading
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_lockable () :
    """
    Test lockable decorator
    """

    # --------------------------------------------------------------------------
    @ru.Lockable
    class Test (object) :

        # ----------------------
        def __init__ (self) :

            self.val = False


        # ----------------------
        def test (self) :

            t.lock ()         # lock before spawning thread
            thread_1 = threading.Thread (target=self.test_1)
            thread_1.start () # thread will run until lock check
            time.sleep (0.1)  # enough time to trigger lock violation
            self.val = False  # set a bogus value
            t.unlock ()       # only now thread can set True
            thread_1.join ()  # make sure the value was set
            assert (self.val) # make sure the value was set correctly

            t.lock ()         # lock before spawning thread
            thread_2 = threading.Thread (target=self.test_2)
            thread_2.start () # thread will run until lock check
            time.sleep (0.1)  # enough time to trigger lock violation
            self.val = False  # set a bogus value
            t.unlock ()       # only now thread can set True
            thread_2.join ()  # make sure the value was set
            assert (self.val) # make sure the value was set correctly


        # ----------------------
        def test_1 (self) :

            with self :
                self.val = True


        # ----------------------
        def test_2 (self) :

            self.lock ()
            self.val = True
            self.unlock ()


    # --------------------------------------------------------------------------
    t = Test ()

    # check lock with resource manager
    with t : 
        pass
    
    # check explicit and recursive lock/unlock
    t.lock   (); assert (    t.locked())
    t.unlock (); assert (not t.locked())
    t.lock   (); assert (    t.locked())
    t.lock   (); assert (    t.locked())
    t.unlock (); assert (    t.locked())
    t.unlock (); assert (not t.locked())

    # check locking over threads
    t.test ()


    # --------------------------------------------------------------------------
    # check double unlock
    try                      : t.unlock (); assert (not t.locked())
    except RuntimeError as e : pass 
    except Exception    as e : assert (False), "RuntimeError != %s" % type(e)
    else                     : assert (False), "expected RuntimeError Exception, got nothing"


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_lockable ()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2014, RADICAL@Rutgers"
__license__   = "MIT"


import time
import inspect
import functools


""" 
Provides class decorators to time all public class methods 

The class decorator is provided as decorator method `@timed_class`, and as
decorator class `@TimedClass`.  It thus also serves as documentation on the
different ways to implement class decorators.  
"""


# ------------------------------------------------------------------------------
#
# called on *decorator class* instantiation
def timed_method (func) :
    """ 
    This class decorator will decorate all public class methods with a timing
    function.  That will time the call invocation, and store the respective data
    in a private class dict '__timing'.  Additionally, it will add the class
    method '_timing_last()', which will return the tuple `[method name, method
    timer]` for the last timed invocation of a class method.
    """

    # apply timing decorator to all public methods
    def func_timer(obj, *args, **kwargs):
        
        try  :
            _ = obj.__timing

        except Exception :

            # never seen this one before -- create private timing dict, and add
            # timing_last method
            obj.__timing = dict ()

            def timing_last () :
                last_call = obj.__timing.get ('__last_call', None)
                timer     = obj.__timing.get (last_call, [None])[0]
                return last_call, timer

            def timing_stats () :
                import math
                ret = ""
                for key in sorted (obj.__timing.keys ()) :

                    if  key == '__last_call' :
                        continue

                    tlist = obj.__timing[key]
                    tnum  = len(tlist)
                    tmean = sum(tlist) / tnum
                    tdev  = [x - tmean for x in tlist]
                    tdev2 = [x*x for x in tdev]
                    tstd  = math.sqrt (sum(tdev2) / tnum)

                    ret  += "%-20s : %10.3fs (+/- %5.3fs) [n=%5d]\n" % (key, tmean, tstd, tnum)

                return ret


            def timing_reset () :
                obj.__timing = dict()

            obj._timing_last  = timing_last
            obj._timing_stats = timing_stats
            obj._timing_reset = timing_reset


        # object is set up -- time the call and record timings
        fname = func.__name__
        tdict = obj.__timing
        
        start = time.time()
        ret   = func (obj, *args, **kwargs)
        stop  = time.time()
        timed = stop-start
        
        if not fname in tdict :
            tdict[fname] = list()
        tdict[fname].append (timed)
        tdict['__last_call'] = fname
        
        return ret
        
    return func_timer






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


"""Parse (absolute and relative) URLs.

See RFC 1808: "Relative Uniform Resource Locators", by R. Fielding,
UC Irvine, June 1995.
"""

__all__ = ["urlparse", "urlunparse", "urljoin", "urldefrag",
           "urlsplit", "urlunsplit"]

# A classification of schemes ('' means apply by default)
uses_relative = ['ftp', 'http', 'gopher', 'nntp', 'imap',
                 'wais', 'file', 'https', 'shttp', 'mms',
                 'prospero', 'rtsp', 'rtspu', '', 'sftp']
uses_netloc = ['ftp', 'http', 'gopher', 'nntp', 'telnet',
               'imap', 'wais', 'file', 'mms', 'https', 'shttp',
               'snews', 'prospero', 'rtsp', 'rtspu', 'rsync', '',
               'svn', 'svn+ssh', 'sftp', 'pbs', 'pbs+ssh', 'sge', 'sge+ssh',
               'fork', 'ssh', 'xt5torque', 'ec2', 'euca', 'pbs+gsissh', 'sge+gsissh',
               'xt5torque+gsissh', 'xt5torque+ssh', 'redis']
non_hierarchical = ['gopher', 'hdl', 'mailto', 'news',
                    'telnet', 'wais', 'imap', 'snews', 'sip', 'sips']
uses_params = ['ftp', 'hdl', 'prospero', 'http', 'imap',
               'https', 'shttp', 'rtsp', 'rtspu', 'sip', 'sips',
               'mms', '', 'sftp']
uses_query = ['http', 'wais', 'imap', 'https', 'shttp', 'mms',
              'gopher', 'rtsp', 'rtspu', 'sip', 'sips', 'srm', '']
uses_fragment = ['ftp', 'hdl', 'http', 'gopher', 'news',
                 'nntp', 'wais', 'https', 'shttp', 'snews',
                 'file', 'prospero', '']
uses_hash_in_hostname = ['go']
preserve_case_in_hostname = ['go']

# Characters valid in scheme names
scheme_chars = ('abcdefghijklmnopqrstuvwxyz'
                'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                '0123456789'
                '+-.')

MAX_CACHE_SIZE = 20
_parse_cache = {}

def clear_cache():
    """Clear the parse cache."""
    global _parse_cache
    _parse_cache = {}


class BaseResult(tuple):
    """Base class for the parsed result objects.

    This provides the attributes shared by the two derived result
    objects as read-only properties.  The derived classes are
    responsible for checking the right number of arguments were
    supplied to the constructor.

    """

    __slots__ = ()

    # Attributes that access the basic components of the URL:

    @property
    def scheme(self):
        return self[0] or ''

    @property
    def netloc(self):
        return self[1]

    @property
    def path(self):
        return self[2] or ''

    @property
    def query(self):
        return self[-2] or ''

    @property
    def fragment(self):
        return self[-1] or ''

    # Additional attributes that provide access to parsed-out portions
    # of the netloc:

    @property
    def username(self):
        netloc = self.netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                userinfo = userinfo.split(":", 1)[0]
            return userinfo
        return ''

    @property
    def password(self):
        netloc = self.netloc
        if "@" in netloc:
            userinfo = netloc.split("@", 1)[0]
            if ":" in userinfo:
                return userinfo.split(":", 1)[1]
        return ''

    @property
    def hostname(self):
        netloc = self.netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            netloc = netloc.split(":", 1)[0]
        if self.scheme in preserve_case_in_hostname:
            return netloc or ''
        else:
            return netloc.lower() or ''

    @property
    def port(self):
        netloc = self.netloc
        if "@" in netloc:
            netloc = netloc.split("@", 1)[1]
        if ":" in netloc:
            port = netloc.split(":", 1)[1]
            return int(port, 10)
        return None


class SplitResult(BaseResult):

    __slots__ = ()

    def __new__(cls, scheme, netloc, path, query, fragment):
        return BaseResult.__new__(
            cls, (scheme, netloc, path, query, fragment))

    def geturl(self):
        return urlunsplit(self)


class ParseResult(BaseResult):

    __slots__ = ()

    def __new__(cls, scheme, netloc, path, params, query, fragment):
        return BaseResult.__new__(
            cls, (scheme, netloc, path, params, query, fragment))

    @property
    def params(self):
        return self[3]

    def geturl(self):
        return urlunparse(self)


def urlparse(url, scheme='', allow_fragments=True):
    """Parse a URL into 6 components:
    <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
    Return a 6-tuple: (scheme, netloc, path, params, query, fragment).
    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes."""
    tuple = urlsplit(url, scheme, allow_fragments)
    scheme, netloc, url, query, fragment = tuple
    if scheme in uses_params and ';' in url:
        url, params = _splitparams(url)
    else:
        params = ''
    return ParseResult(scheme, netloc, url, params, query, fragment)

def _splitparams(url):
    if '/'  in url:
        i = url.find(';', url.rfind('/'))
        if i < 0:
            return url, ''
    else:
        i = url.find(';')
    return url[:i], url[i+1:]

def _splitnetloc(url, start=0, allow_hash_in_hostname=False):
    delim = len(url)   # position of end of domain part of url, default is end
    if allow_hash_in_hostname:
        separators = '/?'
    else:
        separators = '/?#'
    for c in separators:    # look for delimiters; the order is NOT important
        wdelim = url.find(c, start)        # find first of this delim
        if wdelim >= 0:                    # if found
            delim = min(delim, wdelim)     # use earliest delim position
    return url[start:delim], url[delim:]   # return (domain, rest)

def urlsplit(url, scheme='', allow_fragments=True):
    """Parse a URL into 5 components:
    <scheme>://<netloc>/<path>?<query>#<fragment>
    Return a 5-tuple: (scheme, netloc, path, query, fragment).
    Note that we don't break the components up in smaller bits
    (e.g. netloc is a single string) and we don't expand % escapes."""
    allow_fragments = bool(allow_fragments)
    key = url, scheme, allow_fragments, type(url), type(scheme)
    cached = _parse_cache.get(key, None)
    if cached:
        return cached
    if len(_parse_cache) >= MAX_CACHE_SIZE: # avoid runaway growth
        clear_cache()
    netloc = query = fragment = ''
    i = url.find(':')
    if i > 0:
        if url[:i] == 'http': # optimize the common case
            scheme = url[:i].lower()
            url = url[i+1:]
            if url[:2] == '//':
                netloc, url = _splitnetloc(url, 2)
            if allow_fragments and '#' in url:
                url, fragment = url.split('#', 1)
            if '?' in url:
                url, query = url.split('?', 1)
            v = SplitResult(scheme, netloc, url, query, fragment)
            _parse_cache[key] = v
            return v
        for c in url[:i]:
            if c not in scheme_chars:
                break
        else:
            scheme, url = url[:i].lower(), url[i+1:]
    if url[:2] == '//':
        if scheme in uses_hash_in_hostname:
            netloc, url = _splitnetloc(url, 2, True)
        else:
            netloc, url = _splitnetloc(url, 2)
    if allow_fragments and scheme in uses_fragment and '#' in url:
        url, fragment = url.split('#', 1)
    if scheme in uses_query and '?' in url:
        url, query = url.split('?', 1)
    v = SplitResult(scheme, netloc, url, query, fragment)
    _parse_cache[key] = v
    return v

def urlunparse((scheme, netloc, url, params, query, fragment)):
    """Put a parsed URL back together again.  This may result in a
    slightly different, but equivalent URL, if the URL that was parsed
    originally had redundant delimiters, e.g. a ? with an empty query
    (the draft states that these are equivalent)."""
    if params:
        url = "%s;%s" % (url, params)
    return urlunsplit((scheme, netloc, url, query, fragment))

def urlunsplit((scheme, netloc, url, query, fragment)):
    if netloc or (scheme and url[:2] != '//'):
        if url and url[:1] != '/': url = '/' + url
        url = '//' + (netloc or '') + url
    if scheme:
        url = scheme + ':' + url
    if query:
        url = url + '?' + query
    if fragment:
        url = url + '#' + fragment
    return url

def urljoin(base, url, allow_fragments=True):
    """Join a base URL and a possibly relative URL to form an absolute
    interpretation of the latter."""
    if not base:
        return url
    if not url:
        return base
    bscheme, bnetloc, bpath, bparams, bquery, bfragment = \
            urlparse(base, '', allow_fragments)
    scheme, netloc, path, params, query, fragment = \
            urlparse(url, bscheme, allow_fragments)
    if scheme != bscheme or scheme not in uses_relative:
        return url
    #if scheme in uses_netloc:
    if netloc:
        return urlunparse((scheme, netloc, path,
                               params, query, fragment))
    netloc = bnetloc
    if path[:1] == '/':
        return urlunparse((scheme, netloc, path,
                           params, query, fragment))
    if not (path or params or query):
        return urlunparse((scheme, netloc, bpath,
                           bparams, bquery, fragment))
    segments = bpath.split('/')[:-1] + path.split('/')
    # XXX The stuff below is bogus in various ways...
    if segments[-1] == '.':
        segments[-1] = ''
    while '.' in segments:
        segments.remove('.')
    while 1:
        i = 1
        n = len(segments) - 1
        while i < n:
            if (segments[i] == '..'
                and segments[i-1] not in ('', '..')):
                del segments[i-1:i+1]
                break
            i = i+1
        else:
            break
    if segments == ['', '..']:
        segments[-1] = ''
    elif len(segments) >= 2 and segments[-1] == '..':
        segments[-2:] = ['']
    return urlunparse((scheme, netloc, '/'.join(segments),
                       params, query, fragment))

def urldefrag(url):
    """Removes any existing fragment from URL.

    Returns a tuple of the defragmented URL and the fragment.  If
    the URL contained no fragments, the second element is the
    empty string.
    """
    if '#' in url:
        s, n, p, a, q, frag = urlparse(url)
        defrag = urlunparse((s, n, p, a, q, ''))
        return defrag, frag
    else:
        return url, ''


test_input = """
      http://a/b/c/d

      g:h        = <URL:g:h>
      http:g     = <URL:http://a/b/c/g>
      http:      = <URL:http://a/b/c/d>
      g          = <URL:http://a/b/c/g>
      ./g        = <URL:http://a/b/c/g>
      g/         = <URL:http://a/b/c/g/>
      /g         = <URL:http://a/g>
      //g        = <URL:http://g>
      ?y         = <URL:http://a/b/c/d?y>
      g?y        = <URL:http://a/b/c/g?y>
      g?y/./x    = <URL:http://a/b/c/g?y/./x>
      .          = <URL:http://a/b/c/>
      ./         = <URL:http://a/b/c/>
      ..         = <URL:http://a/b/>
      ../        = <URL:http://a/b/>
      ../g       = <URL:http://a/b/g>
      ../..      = <URL:http://a/>
      ../../g    = <URL:http://a/g>
      ../../../g = <URL:http://a/../g>
      ./../g     = <URL:http://a/b/g>
      ./g/.      = <URL:http://a/b/c/g/>
      /./g       = <URL:http://a/./g>
      g/./h      = <URL:http://a/b/c/g/h>
      g/../h     = <URL:http://a/b/c/h>
      http:g     = <URL:http://a/b/c/g>
      http:      = <URL:http://a/b/c/d>
      http:?y         = <URL:http://a/b/c/d?y>
      http:g?y        = <URL:http://a/b/c/g?y>
      http:g?y/./x    = <URL:http://a/b/c/g?y/./x>
"""

def test():
    import sys
    base = ''
    if sys.argv[1:]:
        fn = sys.argv[1]
        if fn == '-':
            fp = sys.stdin
        else:
            fp = open(fn)
    else:
        try:
            from cStringIO import StringIO
        except ImportError:
            from StringIO import StringIO
        fp = StringIO(test_input)
    while 1:
        line = fp.readline()
        if not line: break
        words = line.split()
        if not words:
            continue
        url = words[0]
        parts = urlparse(url)
        print '%-10s : %s' % (url, parts)
        abs = urljoin(base, url)
        if not base:
            base = abs
        wrapped = '<URL:%s>' % abs
        print '%-10s = %s' % (url, wrapped)
        if len(words) == 3 and words[1] == '=':
            if wrapped != words[2]:
                print 'EXPECTED', words[2], '!!!!!!!!!!'

if __name__ == '__main__':
    test()








import sys
import os
import time
import atexit
import signal
import multiprocessing

# from 
# http://stackoverflow.com/questions/1417631/python-code-to-daemonize-a-process


class Daemon (object) :
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, pidfile=None, 
                  stdin='/dev/null', 
                  stdout='/dev/null', 
                  stderr='/dev/null'):

        self.stdin   = stdin
        self.stdout  = stdout
        self.stderr  = stderr
        self.pidfile = pidfile
        self.pid     = None
        self.queue   = multiprocessing.Queue ()


    # --------------------------------------------------------------------------
    #
    def daemonize (self, debug=False) :
        """
        do the UNIX double-fork magic, see Stevens' "Advanced 
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """

        try :

            # first fork
            try : 
                f1_pid = os.fork () 
                if  f1_pid > 0:
                    # wait for daemon pid from second parent
                    self.pid = self.queue.get ()

                    # we are done...
                    return False

            except OSError as e : 
                raise RuntimeError ("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))


            # decouple from parent environment
            os.chdir  ("/") 
          # os.setsid () 
            os.umask  (0) 

            # second fork
            try : 
                f2_pid = os.fork () 
                
                if  f2_pid > 0 :
                    # communicate pid to first parent
                    self.queue.put (f2_pid)

                    # exit from second parent
                    sys.exit(0) 

            except OSError as e : 
                # no use rasing exceptions at this point
                sys.stderr.write ("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
                sys.exit (1) 

            if  not debug :
                # redirect standard file descriptors
                sys.stdout.flush ()
                sys.stderr.flush ()
                
                si = file (self.stdin,  'r' )
                so = file (self.stdout, 'a+')
                se = file (self.stderr, 'a+', 0)
                
                os.dup2 (si.fileno (), sys.stdin.fileno  ())
                os.dup2 (so.fileno (), sys.stdout.fileno ())
                os.dup2 (se.fileno (), sys.stderr.fileno ())

            # write pidfile
            if  self.pidfile :
                atexit.register (self.delpid)
                pid = str (os.getpid ())
                file (self.pidfile,'w+').write ("%s\n" % pid)

            return True

        except Exception as e :
            raise


    # --------------------------------------------------------------------------
    #
    def delpid (self) :

        if  self.pidfile :
            os.remove(self.pidfile)


    # --------------------------------------------------------------------------
    #
    def start (self, debug=False) :

      # if  debug :
      #     return self.run ()

        if  self.pidfile :
            
            # Check for a pidfile to see if the daemon already runs
            try :
                pf  = file (self.pidfile,'r')
                pid = int  (pf.read ().strip ())
                pf.close ()
            except IOError :
                pid = None

            if pid :
                raise RuntimeError ("pidfile %s exist - daemon running?\n" \
                                   % self.pidfile)

        # Start the daemon, and in the demon, run the workload
        if self.daemonize (debug=debug) :
            try :
                self.run ()
            except Exception as e :
                # FIXME: we need a logfile :P
                pass


    # --------------------------------------------------------------------------
    #
    def stop (self) :
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        if  self.pidfile :
            try :
                pf  = file (self.pidfile,'r')
                pid = int  (pf.read ().strip ())
                pf.close ()
            except IOError :
                pid = None

        else :
            pid = self.pid


        if  not pid:
            raise RuntimeError ("no pidfile %s / pid not known.  Daemon not running?\n" \
                               % self.pidfile)

        # Try killing the daemon process    
        try:
            while 1:
                os.kill (pid, signal.SIGTERM)
                time.sleep (0.1)

        except OSError as err :
            err = str(err)

            if  err.find ("No such process") :
                if  os.path.exists (self.pidfile) :
                    os.remove (self.pidfile)

            else :
                raise

    # --------------------------------------------------------------------------
    #
    def restart (self) :
        """
        Restart the daemon
        """
        self.stop  ()
        self.start ()


    # --------------------------------------------------------------------------
    #
    def run (self) :
        """
        You should override this method when you subclass Daemon. 
        It will be called after the process has been
        daemonized by start() or restart().
        """
        raise RuntimeError ("deamon workload undefined")






import re
import os
import sys

_pat = '^\s*(?P<detail>(?P<short>[^-]*)(?P<base>-[^-@]+?)?(-[^@]+?)?(?P<branch>@.+?)?)\s*$'

# ------------------------------------------------------------------------------
#
# versioning mechanism:
#
#   - version_short :  1.2.3                       - used for installation
#   - version_detail:  1.2.3-v1.1-9-g0684b06@devel - used at runtime (ru.version)
#   - both are derived from the last git tag and branch information
#   - VERSION files are created on install, with the version_detail
#
def get_version (paths=None):
    """
    paths:
        a VERSION file containing the detailed version is checked for in every
        directory listed in paths.   When we find a VERSION file, we also look
        for an SDIST file, and return the found name and location as absolute
        path to the sdist.
    """

    if not paths:
        # by default, get version for myself
        pwd   = os.path.dirname (__file__)
        root  = "%s/.." % pwd
        paths = [root, pwd]
    
    if not isinstance (paths, list):
        paths = [paths]
    
    version_short  = None
    version_detail = None
    version_base   = None
    version_branch = None
    sdist_name     = None
    sdist_path     = None
    err            = ''
    
    
    
    # if in any of the paths a VERSION file exists, we use the detailed version
    # in there.
    for path in paths:

        try:
            version_path = "%s/VERSION" % path
    
            with open (version_path) as f:
                line = f.readline()
                line.strip()
                pattern = re.compile (_pat)
                match   = pattern.search (line)
    
                if match:
                    version_short  = match.group ('short').strip()
                    version_detail = match.group ('detail').strip()
                    version_base   = match.group ('base').strip()
                    version_branch = match.group ('branch').strip()
                  # print 'reading        : %s' % version_path
                  # print '               : %s' % line
                  # print 'version_base   : %s' % version_base
                  # print 'version_detail : %s' % version_detail
                  # print 'version_short  : %s' % version_short
                  # print 'version_branch : %s' % version_branch
                    break
    
        except Exception as e:
            # ignore missing VERSION file -- this is caught below.  But ew keep
            # the error message
            err += '%s\n' % repr(e)
            pass


    if version_detail:
        # check if there is also an SDIST near the version_path
        sdist_path = version_path.replace ('/VERSION', '/SDIST')
        try:
            sdist_name = open(sdist_path).read().strip()
        except Exception as e:
            # ignore missing SDIST file
            pass
        
        sdist_path = version_path.replace ('/VERSION', '/%s' % sdist_name)

    # check if any one worked ok
    if version_detail:
        return (version_short, version_detail, version_base, \
                version_branch, sdist_name, sdist_path)
    else:
        raise RuntimeError ("Cannot determine version from %s (%s)" % 
                            (paths, err.strip()))


# ------------------------------------------------------------------------------




__author__    = "Radical.Utils Development Team (Mark Santcroos)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


import time
import sys
import threading
import radical.utils as ru

SIZE = 10
MAX  = 4

def debug(arg):
  # print arg
    pass
# ------------------------------------------------------------------------------
#
def test_lease_manager() :
    """
    Test LeaseManager
    """

    # --------------------------------------------------------------------------
    class Test (object) :

        # ----------------------
        def __init__ (self) :

            self.val = False


        # ----------------------
        def test(self) :

            self.lm = ru.LeaseManager(SIZE)
          # self.lm2 = ru.LeaseManager()

            iw_thread_1 = threading.Thread(target=self.iw_thread, kwargs={'id': 1, 'pool': 'pool1'})
            iw_thread_1.start() # thread will run until lock check

            iw_thread_2 = threading.Thread(target=self.iw_thread, kwargs={'id': 2, 'pool': 'pool2'})
            iw_thread_2.start() # thread will run until lock check

            ow_thread_1 = threading.Thread(target=self.ow_thread, kwargs={'id': 1, 'pool': 'pool1'})
            ow_thread_1.start() # thread will run until lock check

            ow_thread_2 = threading.Thread(target=self.ow_thread, kwargs={'id': 2, 'pool': 'pool2'})
            ow_thread_2.start() # thread will run until lock check

            mon_thread = threading.Thread(target=self.mon_thread, kwargs={'id': 1, 'pool': 'pool1'})
            mon_thread.start() # thread will run until lock check

            # mon_thread.join()
            # iw_thread_1.join()
            # iw_thread_2.join()
            # ow_thread_1.join()
            # ow_thread_2.join()

        # -----------------
        def iw_thread(self, id, pool):
            name = "IWo Thread-%d (%s)" % (id, pool)

            while True:
                leases = list()
                for i in range(MAX):
                    debug("%s: >" % name)
                    lease = self.lm.lease(pool, dict)
                    debug("%s:  =" % name)
                    lease.obj['name'] = name
                    leases.append(lease)

              # time.sleep(.1)

                for lease in leases:
                    #self.lm.release('lease')
                    self.lm.release(lease)
                    #self.lm.release(lease, delete=True)
                    #self.lm2.release(lease)
                    debug("%s:   <" % name)


        # ----------------------
        def ow_thread(self, id, pool) :

            name = "OWo Thread-%d (%s)" % (id, pool)

            while True:
                leases = list()
                for i in range(MAX):
                    debug("%s: >" % name)
                    lease = self.lm.lease(pool, dict)
                    debug("%s:  =" % name)
                    lease.obj['name'] = name
                    leases.append(lease)

              # time.sleep(.1)

                for lease in leases:
                    self.lm.release(lease)
                    debug("%s:   <" % name)


        # ----------------------
        def mon_thread(self, id, pool) :

            name = "Mon Thread-%d (%s)" % (id, pool)

            while True:
                leases = list()
                for i in range(MAX):
                    debug("%s: >" % name)
                    lease = self.lm.lease(pool, dict)
                    debug("%s:  =" % name)
                    lease.obj['name'] = name
                    leases.append(lease)

              # time.sleep(.1)

                for lease in leases:
                    self.lm.release(lease)
                    debug("%s:   <" % name)
                    #sys.exit(1)


    # --------------------------------------------------------------------------
    t = Test()

    t.test()


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_lease_manager()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import threading 


def Lockable (cls) :
    """ 
    This class decorator will add lock/unlock methods to the thusly decorated
    classes, which will be enacted via an also added `threading.RLock` member
    (`self._rlock`)::

        @Lockable
        class A (object) :
        
            def call (self) :
                print 'locked: %s' % self._locked
        
    The class instance can then be used like this::

        a = A    ()
        a.call   ()
        a.lock   ()
        a.call   ()
        a.lock   ()
        a.call   ()
        a.unlock ()
        a.call   ()
        with a :
          a.call ()
        a.call   ()
        a.unlock ()
        a.call   ()
    
    which will result in::

        locked: 0
        locked: 1
        locked: 2
        locked: 1
        locked: 2
        locked: 1
        locked: 0

    The class A can also internally use the lock, and can, for example, use:

        @Lockable
        class A (object) :
            ...
            def work (self) :
                with self :
                    # locked code section
                    ...
    """

    if  hasattr (cls, '__enter__') :
        raise RuntimeError ("Cannot make '%s' lockable -- has __enter__" % cls)

    if  hasattr (cls, '__exit__') :
        raise RuntimeError ("Cannot make '%s' lockable -- has __exit__" % cls)

    if  hasattr (cls, '_rlock') :
        raise RuntimeError ("Cannot make '%s' lockable -- has _rlock" % cls)

    if  hasattr(cls, '_locked') :
        raise RuntimeError ("Cannot make '%s' lockable -- has _locked" % cls)

    if  hasattr(cls, 'locked') :
        raise RuntimeError ("Cannot make '%s' lockable -- has locked" % cls)

    if  hasattr (cls, 'lock') :
        raise RuntimeError ("Cannot make '%s' lockable -- has lock()" % cls)

    if  hasattr (cls, 'unlock') :
        raise RuntimeError ("Cannot make '%s' lockable -- has unlock()" % cls)


    def locked      (self)        : return self._locked
    def locker      (self)        : self._rlock.acquire (); self._locked += 1
    def unlocker    (self, *args) : self._rlock.release (); self._locked -= 1

    cls._rlock    = threading.RLock ()
    cls._locked   = 0
    cls.locked    = locked
    cls.is_locked = locked
    cls.lock      = locker
    cls.unlock    = unlocker
    cls.__enter__ = locker
    cls.__exit__  = unlocker

    return cls


# ------------------------------------------------------------------------------

# @Lockable
# class A (object) :
# 
#     def call (self) :
#         print 'locked 1: %s' % self.locked ()
# 
#         with self :
#             print 'locked 2: %s' % self.locked ()
# 
#         print 'locked 3: %s\n' % self.locked ()
# 
# print
# a = A()
# a.call ()
# a.lock ()
# a.call ()
# a.lock ()
# a.call ()
# a.unlock ()
# a.call ()
# with a :
#   a.call ()
# a.call ()
# a.unlock ()
# a.call ()
# try :
#     a.unlock ()
#     print 'oops\n'
# except :
#     print 'ok\n'
# a.lock ()
# a.call ()
# with a :
#   a.call ()
# a.call ()
# a.unlock ()
# a.call ()
# try :
#     a.unlock ()
#     print 'oops\n'
# except :
#     print 'ok\n'

# ------------------------------------------------------------------------------






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import time
import uuid
import fcntl
import socket
import datetime
import threading
import singleton

from .misc import dockerized, get_radical_base


# ------------------------------------------------------------------------------
#
class _IDRegistry(object):
    """
    This helper class (which is not exposed to any user of radical.utils)
    generates a sequence of continous numbers for each known ID prefix.  It is
    a singleton, and thread safe (assuming that the Singleton metaclass supports
    thread safe construction).
    """

    __metaclass__ = singleton.Singleton


    # --------------------------------------------------------------------------
    def __init__(self):
        """
        Initialized the registry dict and the threading lock
        """

        self._rlock    = threading.RLock()
        self._registry = dict()


    # --------------------------------------------------------------------------
    def get_counter(self, prefix):
        """
        Obtain the next number in the sequence for the given prefix.
        If the prefix is not known, a new registry counter is created.
        """

        with self._rlock:

            if not prefix in self._registry:
                self._registry[prefix] = 0

            ret = self._registry[prefix]

            self._registry[prefix] += 1

            return ret


    # --------------------------------------------------------------------------
    def reset_counter(self, prefix, reset_all_others=False):
        """
        Reset the given counter to zero.
        """

        with self._rlock:
            
            if reset_all_others:
                # reset all counters *but* the one given
                for p in self._registry:
                    if p != prefix:
                        self._registry[p] = 0
            else:
                self._registry[prefix] = 0


# ------------------------------------------------------------------------------
#
# we create on private singleton instance for the ID registry.
#
_id_registry = _IDRegistry()
_BASE        = get_radical_base('utils')


# ------------------------------------------------------------------------------
#
ID_SIMPLE  = 'simple'
ID_UNIQUE  = 'unique'
ID_PRIVATE = 'private'
ID_CUSTOM  = 'custom'
ID_UUID    = 'uiud'

# ------------------------------------------------------------------------------
#
def generate_id(prefix, mode=ID_SIMPLE, namespace=None):
    """
    Generate a human readable, sequential ID for the given prefix.

    The ID is by default very simple and thus very readable, but cannot be
    assumed to be globally unique -- simple ID uniqueness is only guaranteed
    within the scope of one python instance.

    If `mode` is set to the non-default type `ID_UNIQUE`, an attempt is made to
    generate readable but globally unique IDs -- although the level of
    confidence for uniqueness is significantly smaller than for, say UUIDs.

    Examples::

        print radical.utils.generate_id('item.')
        print radical.utils.generate_id('item.')
        print radical.utils.generate_id('item.', mode=radical.utils.ID_SIMPLE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_SIMPLE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_UNIQUE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_UNIQUE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_PRIVATE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_PRIVATE)
        print radical.utils.generate_id('item.', mode=radical.utils.ID_UUID)

    The above will generate the IDs:

        item.0001
        item.0002
        item.0003
        item.0004
        item.2014.07.30.13.13.44.0001
        item.2014.07.30.13.13.44.0002
        item.cameo.merzky.021342.0001
        item.cameo.merzky.021342.0002
        item.23cacb7e-0b08-11e5-9f0f-08002716eaa9

    where 'cameo' is the (short) hostname, 'merzky' is the username, and '02134'
    is 'days since epoch'.  The last element, the counter is unique for each id
    type and item type, and restarts for each session (application process).  In
    the last case though (`ID_PRIVATE`), the counter is reset for every new day,
    and can thus span multiple applications.

    'namespace' argument can be specified to a value such that unique IDs are 
    created local to that namespace, . For example, you can create a session
    and use the session ID as a namespace for all the IDs of the objects of that
    execution. 

    Example:: 

        sid   = ru.generate_id('re.session', ru.ID_PRIVATE)
        print sid, ru.generate_id('pipeline.%(item_counter)04d', ru.ID_CUSTOM, namespace=sid)
        print sid, ru.generate_id('pipeline.%(item_counter)04d', ru.ID_CUSTOM, namespace=sid)


    This will generate the following output::

        re.session.vivek-HP-Pavilion-m6-Notebook-PC.vivek.017548.0001 pipeline.0000
        re.session.vivek-HP-Pavilion-m6-Notebook-PC.vivek.017548.0001 pipeline.0001

    The namespaces are stored under ```$RADICAL_BASE_DIR/.radical/utils/```.  If
    `RADICAL_BASE_DIR` is not set, then `$HOME` is used.

    Note that for docker containers, we try to avoid hostname / username clashes
    and will, for `ID_PRIVATE`, revert to `ID_UUID`.
    """

    if not prefix or \
        not isinstance(prefix, basestring):
        raise TypeError("ID generation expect prefix in basestring type")

    template = ""

    if dockerized() and mode == ID_PRIVATE:
         mode = ID_UUID

    if   mode == ID_SIMPLE : template = "%(prefix)s.%(counter)04d"
    elif mode == ID_UNIQUE : template = "%(prefix)s.%(date)s.%(time)s.%(pid)06d.%(counter)04d"
    elif mode == ID_PRIVATE: template = "%(prefix)s.%(host)s.%(user)s.%(days)06d.%(day_counter)04d"
    elif mode == ID_CUSTOM : template = prefix
    elif mode == ID_UUID   : template = "%(prefix)s.%(uuid)s"
    else:
        raise ValueError("mode '%s' not supported for ID generation", mode)

    return _generate_id(template, prefix, namespace)

# ------------------------------------------------------------------------------
#
def _generate_id(template, prefix, namespace=None):

    # FIXME: several of the vars below are constants, and many of them are
    # rarely used in IDs.  They should be created only once per module instance,
    # and/or only if needed.

    import getpass

    state_dir = _BASE
    if namespace:
        state_dir += '/%s' % namespace

    if not os.path.isdir(state_dir):
        os.makedirs(state_dir)

    # seconds since epoch(float), and timestamp
    seconds = time.time()
    now     = datetime.datetime.fromtimestamp(seconds)
    days    = int(seconds / (60*60*24))

    try:
        user = getpass.getuser()
    except Exception:
        user = 'nobody'

    info = dict()

    info['day_counter' ]  = 0
    info['item_counter']  = 0
    info['counter'     ]  = 0
    info['prefix'      ]  = prefix
    info['now'         ]  = now
    info['seconds'     ]  = int(seconds)              # full seconds since epoch
    info['days'        ]  = days                      # full days since epoch
    info['user'        ]  = user                      # local username
    info['date'        ]  = "%04d.%02d.%02d" % (now.year, now.month,  now.day)
    info['time'        ]  = "%02d.%02d.%02d" % (now.hour, now.minute, now.second)
    info['pid'         ]  = os.getpid()

    # the following ones are time consuming, and only done when needed
    if '%(host)' in template: info['host'] = socket.gethostname() # local hostname
    if '%(uuid)' in template: info['uuid'] = uuid.uuid1()         # pain old uuid

    if '%(day_counter)' in template:
        fd = os.open("%s/ru_%s_%s.cnt" % (state_dir, user, days), os.O_RDWR | os.O_CREAT)
        fcntl.flock(fd, fcntl.LOCK_EX)
        os.lseek(fd, 0, os.SEEK_SET )
        data = os.read(fd, 256)
        if not data: data = 0
        info['day_counter'] = int(data)
        os.lseek(fd, 0, os.SEEK_SET )
        os.write(fd, "%d\n" % (info['day_counter']+1))
        os.close(fd)

    if '%(item_counter)' in template:
        fd = os.open("%s/ru_%s_%s.cnt" % (state_dir, user, prefix), os.O_RDWR | os.O_CREAT)
        fcntl.flock(fd, fcntl.LOCK_EX)
        os.lseek(fd, 0, os.SEEK_SET)
        data = os.read(fd, 256)
        if not data: data = 0
        info['item_counter'] = int(data)
        os.lseek(fd, 0, os.SEEK_SET)
        os.write(fd, "%d\n" % (info['item_counter']+1))
        os.close(fd)

    if '%(counter)' in template:
        info['counter'] = _id_registry.get_counter(prefix.replace('%', ''))


    ret = template % info

    if '%(' in ret:
      # import pprint
      # pprint.pprint(info)
      # print template
      # print ret
        raise ValueError('unknown replacement pattern in template (%s)' % template)

    return ret


# ------------------------------------------------------------------------------
#
def reset_id_counters(prefix=None, reset_all_others=False):

    if not isinstance(prefix, list):
        prefix = [prefix]

    for p in prefix:
        _id_registry.reset_counter(p.replace('%', ''), reset_all_others)


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


''' Provides API handles for configuration systems. '''

import os
import copy
import threading

from ..singleton import Singleton

from   configfile import ConfigFileReader


# ------------------------------------------------------------------------------
#
def getConfig(name):
    """ 
    Returns a handle to the global configuration object.
    """
    return Configuration(name) 


# ------------------------------------------------------------------------------
#
class ConfigOption(object):
    """ 
    Represent a (mutable) configuration option.
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, category, name, val_type, default_val, valid_options,
                 documentation, env_var):

        self._category      = category
        self._name          = name
        self._val_type      = val_type
        self._default_val   = default_val
        self._valid_options = valid_options
        self._env_var       = env_var
        self._documentation = documentation
        self._value         = None


    # --------------------------------------------------------------------------
    #
    def __str__(self):
        return str({'name':self._name, 'value':self._value})


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        return {self._name:self._value}


    # --------------------------------------------------------------------------
    #
    @property
    def category(self):
        return self._category


    # --------------------------------------------------------------------------
    #
    def set_value(self, value):
        # make sure we got the right value type
        if type(value) != self._val_type:
            raise ValueTypeError(self._category, self._name, 
              type(value), self._val_type)

        self._value = value


    # --------------------------------------------------------------------------
    #
    def get_value(self):
        return self._value


# ------------------------------------------------------------------------------
#
# a singleton class which keeps all configurations
#
class _Configurations (object) :
    """
    This singleton class maintains references to all global configurations.
    """
    __metaclass__ = Singleton
    _rlock = threading.RLock ()

    # --------------------------------------------------------------------------
    #
    def __init__ (self) :


        with self._rlock :
            self._configs = {}


    # --------------------------------------------------------------------------
    #
    def put (self, name, config) :

        with self._rlock :
            self._configs[name] = dict (config)
          # self._configs[name] = copy.deepcopy (config)


    # --------------------------------------------------------------------------
    #
    def get (self, name) :

        with self._rlock :
            if  not name in self._configs :
                self._configs[name] = dict()

            return dict (self._configs[name])
          # return copy.deepcopy (self._configs[name])


# ------------------------------------------------------------------------------
#
_configurations = _Configurations ()


# ------------------------------------------------------------------------------
#
class Configuration (object): 
    """ 
    Represents a global configuration for a given namespace.

    The Configuration class can be used to introspect and modify configuration
    options.  It uses a 'Singleton' data store, which means that  multiple
    instances all point to the same global configuration, for a specific given
    namespace.
    """    

    _rlock = threading.RLock ()


    # --------------------------------------------------------------------------
    #
    def __init__ (self, name):

        with self._rlock :

            self._name = name

            # initialize and sync options
            self._initialize()


    # --------------------------------------------------------------------------
    #
    def _put_state (self):

        with self._rlock :

            # push state back into the store
            state = {}
            state['all_valid_options'] = self._all_valid_options
            state['master_config']     = self._master_config

            _configurations.put (self._name, state)


    # --------------------------------------------------------------------------
    #
    def _get_state (self):

        with self._rlock :

            # populate this instance with previously derived config options
            state = _configurations.get (self._name)

            # _all_valid_options contains the raw option sets
            # added to configuration by 'Configurable' classes
            if 'all_valid_options' in state :
                self._all_valid_options = state['all_valid_options']
            else :
                self._all_valid_options = list()

            # _master config holds the parsed options, including 
            # values fetched from configuartion files and env variables
            if 'master_config' in state :
                self._master_config = state['master_config']
            else :
                self._master_config = dict()



    # --------------------------------------------------------------------------
    #
    def _update (self, namespace, valid_options):

        with self._rlock :

            # sync for singletonism
            self._get_state ()

            # add the new options to the global dictionary
            self._all_valid_options += valid_options

            # sync for singletonism
            self._put_state ()

            # re-_initialize the configuration object.
            self._initialize()

    # --------------------------------------------------------------------------
    #
    def _initialize(self, inject_cfg_file=None, add_cfg_file=None):
        """ Initialize the global configuration.

            :param inject_cfg_file: is used *only* for testing purposes 
             and overwrites / ignores the regular config file locations 
             /etc/<name>.cfg & $HOME/.<name>.cfg

            :param add_cfg_file: is used *only* for testing purposes 
             and adds a specific config file to the list of evaluated files.  
        """
        with self._rlock :

            # get state from cache
            self._get_state ()

            cfg_files = list()
            if inject_cfg_file is not None:
                cfg_files.append(inject_cfg_file)
            else:
                # check for the existence of regular configuration files
                sys_cfg = '/etc/%s.cfg' % self._name
                if os.path.exists(sys_cfg):
                    cfg_files.append(sys_cfg)

                uconf = "%s_CONFIG" % self._name.upper ()
                uconf = uconf.replace ('.', '_')
                uconf = uconf.replace ('-', '_')
                if uconf in os.environ :
                    usr_cfg = os.environ[uconf]
                    if  not os.path.exists (usr_cfg) :
                        print "WARNING: %s set to %s, but file does not exist" % (uconf, usr_cfg)
                else:
                    usr_cfg = '%s/.%s.cfg' % (os.path.expanduser("~"), self._name)

                if os.path.exists(usr_cfg):
                    cfg_files.append(usr_cfg)

            if add_cfg_file is not None:
                cfg_files.append(add_cfg_file)

            cfr = ConfigFileReader(cfg_files)
    
            # cfg_file_dict holds all configuration variables that 
            # were read from either a system-wide or user configuration file.
            cfg_file_dict = cfr.get_config_dict()


            # load valid options and add them to the configuration
            for option in self._all_valid_options:
                cat = option['category']
                if cat not in self._master_config:
                    # first occurrence - add new category key
                    self._master_config[cat] = dict()

                ev = None
                if option['env_variable'] :
                    ev = os.environ.get(option['env_variable'])

                if (option['category'] in cfg_file_dict) and \
                    option['name']     in cfg_file_dict[option['category']]:
                    # found entry in configuration file -- use it
                    tmp_value = cfg_file_dict[option['category']][option['name']]
                    # some list types need type conversion
                    if option['type'] == list:
                        value = tmp_value.split(",")
                    elif option['type'] == bool:
                        if tmp_value.lower() == 'true':
                          value = True
                        elif tmp_value.lower() == 'false':
                          value = False 
                        else:
                          raise ValueTypeError(option['category'], option['name'],
                              tmp_value, option['type'])
                    elif option['type'] == int:
                        value = int(tmp_value)
                    elif option['type'] == float:
                        value = float(tmp_value)
                    else:
                        value = str(tmp_value)

                elif ev is not None:
                    tmp_value = ev
                    if option['type'] == float:
                        value = float(tmp_value)
                    elif option['type'] == int:
                        value = int(tmp_value)
                    elif option['type'] == list:
                        value = tmp_value.split(",")
                    elif option['type'] == bool:
                        if tmp_value.lower() == 'true':
                          value = True
                        elif tmp_value.lower() == 'false':
                          value = False 
                        elif tmp_value == '1':
                          value = True
                        elif tmp_value == '0':
                          value = False
                        else:
                          raise ValueTypeError(option['category'], option['name'],
                              tmp_value, option['type'])
                    elif option['type'] == int:
                        value = int(tmp_value)
                    elif option['type'] == float:
                        value = float(tmp_value)
                    else:
                        value = tmp_value
                else:
                    value = option['default']

                if not 'valid_options' in option :
                    option['valid_options'] = None

                self._master_config[cat][option['name']] = ConfigOption(
                    option['category'],
                    option['name'],
                    option['type'],
                    option['default'],
                    option['valid_options'],
                    option['documentation'],
                    option['env_variable'])

                self._master_config[cat][option['name']].set_value(value) 


            # now walk through the cfg_file_dict -- for all entries not yet handled
            # (i.e. which have not been registered before), use default
            # ConfigOptions:
            #       category       : from cfg_file_dict
            #       name           : from cfg_file_dict
            #       value          : from cfg_file_dict
            #       type           : string
            #       default        : ""
            #       valid_options  : None
            #       documentation  : ""
            #       env_variable   : None
            #
            # If later initialize_ is called again, and that option has been
            # registered meanwhile, this entry will be overwritten.

            for section in cfg_file_dict.keys () :
                for name in cfg_file_dict[section].keys () :
                    value = cfg_file_dict[section][name]

                    # check if this is a registered entry
                    exists = False
                    if section in self._master_config :
                        if name in self._master_config[section] :
                            exists = True
                    else :
                        # section does not exist - create it, and add option
                        self._master_config[section] = {}

                    # if not, add it with default ConfigOptions
                    if not exists :
                        self._master_config[section][name] = ConfigOption(
                            section   , # category
                            name      , # name
                            str       , # type
                            ""        , # default
                            None      , # valid_options
                            ""        , # documentation
                            None      ) # env_variable

                        self._master_config[section][name].set_value(value)


            # sync cache with newly configured state
            self._put_state ()

    # --------------------------------------------------------------------------
    #
    def has_category(self, category_name):
        """ Check for a specific configuration category.
        """
        with self._rlock :

            # sync for singletonism
            self._get_state ()

            if category_name not in self._master_config:
                return False
            else:
                return True

    # --------------------------------------------------------------------------
    #
    def get_category(self, category_name):
        """ Return a specific configuration category.
        """
        with self._rlock :

            # sync for singletonism
            self._get_state ()

            if category_name not in self._master_config:
                raise CategoryNotFound(category_name)
            else:
                return self._master_config[category_name]

    # --------------------------------------------------------------------------
    #
    def has_option(self, category_name, option_name):

        with self._rlock :

            # sync for singletonism
            self._get_state ()

            if category_name not in self._master_config:
                return False
            else:
                if option_name not in self._master_config[category_name]:
                    return False
                else:
                    return True


    # --------------------------------------------------------------------------
    #
    def get_option(self, category_name, option_name):

        with self._rlock :

            # sync for singletonism
            self._get_state ()

            if category_name not in self._master_config:
                raise CategoryNotFound(category_name)
            else:
                if option_name not in self._master_config[category_name]:
                    raise OptionNotFound(category_name, option_name)
                else:
                    return self._master_config[category_name][option_name]


    # --------------------------------------------------------------------------
    #
    def as_dict (self, cn = None) :

        with self._rlock :

            # sync for singletonism
            self._get_state ()

            ret = {}

            if  cn : 
                for on in self._master_config[cn] :
                    ret[on] = self.get_option (cn, on).get_value ()

            else :
                for cn in self._master_config :
                    ret[cn] = {}
  
                    for on in self._master_config[cn] :
                        ret[cn][on] = self.get_option (cn, on).get_value ()

            return ret



# ------------------------------------------------------------------------------
#
class Configurable (object) :
    """ 
    This class provides an interface for all configurable objects.  
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, name):

        self._name = name

    # --------------------------------------------------------------------------
    #
    def config_options (self, namespace, valid_options):

        ## register a new 'configurable' object
        getConfig(self._name)._update(namespace, valid_options)


    # --------------------------------------------------------------------------
    #
    def get_config(self, namespace):
        return getConfig(self._name).get_category(namespace)



    # --------------------------------------------------------------------------
    #
    def get_config_as_dict (self): return getConfig(self._name).as_dict ()
    def            as_dict (self): return getConfig(self._name).as_dict ()


# ------------------------------------------------------------------------------
#
class CategoryNotFound(LookupError):
    # --------------------------------------------------------------------------
    #
    def __init__(self, name):
        msg = "A category with name '%s' could not be found." % name
        LookupError.__init__(self, msg)


# ------------------------------------------------------------------------------
#
class OptionNotFound(LookupError):
    # --------------------------------------------------------------------------
    #
    def __init__(self, category_name, option_name):
        name = "%s.%s" % (category_name, option_name)
        msg  = "An option with name '%s' could not be found." % (name)
        LookupError.__init__(self, msg)


# ------------------------------------------------------------------------------
#
class ValueTypeError(TypeError):
    # --------------------------------------------------------------------------
    #
    def __init__(self, category_name, option_name, value_type, required_type):
        name = "%s.%s" % (category_name, option_name)
        msg  = "Option %s requires value of type '%s' but got '%s'." \
             % (name, required_type, value_type)
        TypeError.__init__ (self, msg)


# ------------------------------------------------------------------------------
#






__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import re
import fnmatch

PRESERVE  = 'preserve'
OVERWRITE = 'overwrite'


# see http://code.activestate.com/recipes/117236-dictionary-mixin-framework/

# ------------------------------------------------------------------------------
#
class DictMixin :
    '''
    Mixin defining all dictionary methods for classes that already have
    a minimum dictionary interface including getitem, setitem, delitem,
    and keys.  Based on those methods, the mixin provides the remaining
    interface functionality to make the class look like a fully compliant
    dictionary.
    '''

    # --------------------------------------------------------------------------
    #
    # first level definitions should be implemented by the sub-class
    #
    def __getitem__(self, key):
        raise NotImplementedError

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        raise NotImplementedError    

    def keys(self):
        raise NotImplementedError

    
    # --------------------------------------------------------------------------
    #
    # second level definitions which assume only getitem and keys
    #
    def has_key(self, key):
         return key in self.keys()

    def __iter__(self):
        for k in self.keys():
            yield k


    # --------------------------------------------------------------------------
    #
    # third level uses second level instead of first
    #
    def __contains__(self, key):
        return self.has_key(key)            

    def iteritems(self):
        for k in self:
            yield (k, self[k])


    # --------------------------------------------------------------------------
    #
    # fourth level uses second and third levels instead of first
    #
    def iterkeys(self):
        return self.__iter__()

    def itervalues(self):
        for _, v in self.iteritems():
            yield v

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def clear(self):
        for key in self.keys():
            del self[key]

    def setdefault(self, key, default):
        if key not in self:
            self[key] = default
            return default
        return self[key]

    def popitem(self):
        key = self.keys()[0]
        value = self[key]
        del self[key]
        return (key, value)

    def update(self, other):
        for key in other.keys():
            self[key] = other[key]

    def get(self, key, default=None):
        if key in self:
            return self[key]
        return default

    def __repr__(self):
        return repr(dict(self.items()))


# ------------------------------------------------------------------------------
#
def dict_merge (a, b, policy=None, wildcards=False, logger=None, _path=[]):
    # thanks to 
    # http://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
    """
    This merges two dict in place, modifying the original dict in a.

    Merge Policies :
        None (default) : raise an exception on conflicts
        PRESERVE       : original value in a are preserved, new values 
                         from b are only added where the original value 
                         is None / 0 / ''
        OVERWRITE      : values in a are overwritten by new values from b

    """

    if  a == None : return
    if  b == None : return

    if  not isinstance (a, dict) :
        raise TypeError ("*dict*_merge expects dicts, not '%s'" % type(a))

    if  not isinstance (b, dict) :
        raise TypeError ("*dict*_merge expects dicts, not '%s'" % type(b))

            
    # --------------------------------------------------------------------------
    def merge_key (a, key_a, b, key_b) :

        # need to resolve conflict
        if  isinstance (a[key_a], dict) and isinstance (b[key_b], dict):
            dict_merge (a[key_a], b[key_b], 
                        policy    = policy, 
                        wildcards = wildcards, 
                        logger    = logger, 
                        _path     = _path + [str(key)])
        
        elif (key_a not in a) and (key_b in b):
            a[key_a] = b[key_b] # use b value

        elif (key_a in a) and (key_b not in b):
            pass # keep a value

        elif a[key_a] == b[key_b]:
            pass # same leaf value

        elif (key_a not in a) and (key_b not in b):
            pass # keep no a value

        else:
            if  policy == PRESERVE:
                if  logger :
                    logger.debug ("preserving key %s:%s \t(%s)" % (":".join(_path), key_b, b[key_b]))
                pass # keep original value

            elif policy == OVERWRITE:
                if  logger :
                    logger.debug ("overwriting key %s:%s \t(%s)" % (":".join(_path), key_b, b[key_b]))
                a[key] = b[key] # use new value

            else :
                raise ValueError ('Conflict at %s (%s : %s)' \
                               % ('.'.join(_path + [str(key)]), a[key_a], b[key_b]))
    # --------------------------------------------------------------------------

    # first a clean merge, i.e. no interpretation of wildcards
    for key in b:
        
        if  key in a:

        # need to resolve conflict
            merge_key (a, key, b, key)
        
        else:
            # no conflict - simply add.  Not that this is a potential shallow
            # copy if b[key] is a complex type.
            a[key] = b[key]


    # optionally, check if other merge options are also valid
    for key_b in b:
        if  wildcards :
            if  '*' in key_b :
                pat = re.compile (fnmatch.translate (key_b))
                for key_a in a :
                    if  pat.match (key_a) :
                        merge_key (a, key_a, b, key_b)

    return a

# ------------------------------------------------------------------------------
#
def dict_stringexpand (target, sources=None) :
    """
    This expands dict entries (strings only) with keys from a second dict. For
    example, the dicts::

        target  = {'workdir'  : '/home/%(user)s/', 
                   'resource' : '%(resource)s'}
        sources = {'user'     : 'peer_gynt',
                   'protocol' : 'ssh',
                   'host'     : 'localhost',
                   'resource' : '%(protocol)s://%(host)s/'}

    would result in::
        target = {'workdir'  : '/home/peer_gynt/', 
                  'resource' : 'ssh://localhost'}

    Note that expansion happened twice, for the `resource` tag to be fully
    specified.
    """

    assert (isinstance(target, dict))

    # expand from self, and all given dicts, but only use 
    # first-level primitive types (string, int, float)
    if  sources :
        if  isinstance (sources, dict) :
            sources = [sources]
    else :
        sources = list()

    if  not isinstance (sources, list) :
        raise TypeError ("Need dict as expansion sources, not %s" % type(sources))

    # target must be first source, to avoid cycles (other sources are likely to
    # have *other* info)
    sources.insert (0, target)

    repl_source = dict()
    for source in sources :
        for (key, val) in source.iteritems() :
            if  isinstance (val, basestring) or \
                isinstance (val, int       ) or \
                isinstance (val, float     ) :
                repl_source[key] = val

  # print '---------'
  # print 'source'
  # import pprint
  # pprint.pprint (repl_source)
  # print '---------'
  # print 'target'
  # pprint.pprint (target)
  # print '---------'

    again = True
    while again :
        target, again = _generic_stringexpand (target, repl_source)
        
      # print 'target x'
      # pprint.pprint (target)
      # print '---------'

    return target


# ------------------------------------------------------------------------------
#
def _generic_stringexpand (target, source) :

  # print
  # print 'generic (%s) (%s) (%s)' % (type(target), id(target), target)


    if  isinstance (target, basestring) : 
        return _string_stringexpand (target, source)

    elif  isinstance (target, list) : 
        return _list_stringexpand (target, source)

    elif  isinstance (target, dict) : 
        return _dict_stringexpand (target, source)

    else :
        # ignore other types for now
        return target, False



# ------------------------------------------------------------------------------
#
def _list_stringexpand (target, source) :

    assert (isinstance(target, list))
    assert (isinstance(source, dict))

  # print '> list %s' % target

    all_again = 0
    for (idx, elem) in enumerate(target) :
        target[idx], again = _generic_stringexpand (elem, source)
        all_again += again

  # print '< list %s' % target
    return target, all_again


# ------------------------------------------------------------------------------
#
def _dict_stringexpand (target, source) :

  # print 'dict'

    assert (isinstance(target, dict))
    assert (isinstance(source, dict))

    all_again = 0
    for (key, val) in target.iteritems() :
      # print 'key: %s' % key
        target[key], again = _generic_stringexpand (val, source)
        all_again += again

    return target, all_again


# ------------------------------------------------------------------------------
#
def _string_stringexpand (target, source) :

  # print 'string %s' % target

    assert (isinstance(target, basestring))
    assert (isinstance(source, dict))

    orig = str(target)
    try :
        expanded = target % source

    except KeyError as e:
        # we ignore incomplete expands
      # print e
        return orig, False

    except ValueError as e:
        # we ignore incomplete expands
      # print e
        return orig, False

    # only check for success after success.  Duh!
  # print '>>> orig (%s) expanded (%s)' % (orig, expanded)
    if  orig == expanded : return expanded, False
    else                 : return expanded, True



# ------------------------------------------------------------------------------






__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os

def which(program):
    '''taken from:
        http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
    '''
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)

    if  fpath:
        if  is_exe (program) :
            return program
    else:
        for path in os.environ["PATH"].split (os.pathsep) :
            exe_file = os.path.join (path, program)
            if  is_exe (exe_file) :
                return exe_file

    return None







__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import re
import json


# ------------------------------------------------------------------------------
#
def read_json (filename) :
    """
    Comments in the form of
        # rest of line
    are stripped from json before parsing

    use like this::

        import pprint
        pprint.pprint (read_json (sys.argv[1]))

    """

    with open (filename) as f:

        content = ''

        # weed out comments
        for line in f.readlines () :
            content += re.sub (r'^\s*#.*', '', line)

        return json.loads (content)


# ------------------------------------------------------------------------------
#
def read_json_str (filename) :
    """
    same as read_json, but converts unicode strings to simple strings.
    This could trivially done with YAML, but that introduces a compile
    dependency -- so we convert manually...
    """

    return _unicode_to_strings (read_json (filename))


# ------------------------------------------------------------------------------
#
def write_json (data,  filename) :
    """
    thin wrapper around python's json write, for consistency of interface

    """

    # thanks to
    # http://stackoverflow.com/questions/956867/#13105359
    def _byteify(input):
        if isinstance(input, dict):
            return {_byteify(key):_byteify(value) for key,value in input.iteritems()}
        elif isinstance(input, list):
            return [_byteify(element) for element in input]
        elif isinstance(input, unicode):
            return input.encode('utf-8')
        else:
            return input


    with open (filename, 'w') as f :
        json.dump (_byteify(data), f, sort_keys=True, indent=4, ensure_ascii=False)
        f.write ("\n")


# ------------------------------------------------------------------------------
#
def parse_json (json_str, filter_comments=True) :
    """
    Comments in the form of
        # rest of line
    are stripped from json before parsing
    """

    if not filter_comments :
        return json.loads (json_str)

    else :
        content = ''
        for line in json_str.split ('\n') :
            content += re.sub (r'^\s*#', '', line)
            content += '\n'

        return json.loads (content)


# ------------------------------------------------------------------------------
#
def parse_json_str (json_str) :
    """
    same as parse_json, but converts unicode strings to simple strings
    """

    return _unicode_to_strings (parse_json (json_str))


# ------------------------------------------------------------------------------
#
def _unicode_to_strings (unicode_data) :

    if isinstance (unicode_data, dict) :
        ret = dict ()
        for key, value in unicode_data.iteritems() :
            ret[_unicode_to_strings(key)] = _unicode_to_strings(value) 
        return ret

    elif isinstance (unicode_data, list) :
        ret = list ()
        for element in unicode_data :
            ret.append (_unicode_to_strings (element))
        return ret

    elif isinstance (unicode_data, unicode) :
        return unicode_data.encode ('utf-8')

    else:
        return unicode_data


# ------------------------------------------------------------------------------




__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import radical.utils as ru


# -------------------------------------------------------------------------
#
def test_url_compatibility () :
    """
    Test basic url compatibility
    """

    u1 = ru.Url("ssh://user:pwd@hostname.domain:9999/path")

    assert u1.scheme   == "ssh"
    assert u1.username == "user"
    assert u1.password == "pwd"
    assert u1.host     == "hostname.domain"
    assert u1.port     == int(9999)


# ------------------------------------------------------------------------------
#
def test_url_scheme_issue () :
    """
    Test url schema issues
    """

    u1 = ru.Url("unknownscheme://user:pwd@hostname.domain:9999/path")

    assert u1.scheme   == "unknownscheme"
    assert u1.username == "user"
    assert u1.password == "pwd"
    assert u1.host     == "hostname.domain"
    assert u1.port     == int(9999)

# ------------------------------------------------------------------------------
#
def test_url_issue_49 () : 
    """
    Test url issue #49
    """

    url = ru.Url ("scheme://pass:user@host:123/dir/file?query#fragment")
    url.set_host   ('remote.host.net')
    url.set_scheme ('sftp') 
    url.set_path   ('/tmp/data')
    
    assert str(url) == "sftp://pass:user@remote.host.net:123/tmp/data"

# ------------------------------------------------------------------------------
#
def test_url_issue_61 () : 
    """
    Test url issue #61
    """

    url = ru.Url ("advert://localhost/?dbtype=sqlite3")
    assert url.query == "dbtype=sqlite3"

# ------------------------------------------------------------------------------
#
def test_url_properties () :
    """
    Test url properties
    """

    url = ru.Url("")
    assert str(url)           == ""
 
    url.scheme = "scheme"
    assert str(url)           == "scheme://"
    assert url.get_scheme()   == "scheme"
 
    url.set_scheme("tscheme")
    assert url.get_scheme()   == "tscheme"
 
    url.scheme = "scheme"
    url.host   = "host"
    assert str(url)           == "scheme://host"
    assert url.get_host()     == "host"
    
    url.set_host("thost")
    assert url.get_host()     == "thost"
    
    url.host = "host"
    url.port = 42
    assert str(url)           == "scheme://host:42"
    assert url.get_port()     == 42
 
    url.set_port(43)
    assert url.get_port()     == 43
 
    url.port     = 42
    url.username = "username"
    assert str(url)           == "scheme://username@host:42"
    assert url.get_username() == "username"
    
    url.set_username("tusername")
    assert url.get_username() == "tusername"
    
    url.username = "username"
    url.password = "password"
    assert str(url)           == "scheme://username:password@host:42"
    assert url.get_password() == "password"
    
    url.set_password("tpassword")
    assert url.get_password() == "tpassword"
    
    url.password = "password"
    url.path     = "/path/"
    assert str(url)           == "scheme://username:password@host:42/path/"
    assert url.get_path()     == "/path/"
 
    url.set_path("tpath")
    assert url.get_path()     == "/tpath"
   


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_url_compatibility ()
    test_url_scheme_issue  ()
    test_url_issue_49      ()
    test_url_issue_61      ()
    test_url_properties    ()





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils.signatures as rus


# # ------------------------------------------------------------------------------
# @rus.takes   (basestring, int, rus.optional (float))
# @rus.returns (int)
# def sigtest (string, intger, float=3.1415926) :
#     return 1
# 
# # ------------------------------------------------------------------------------
# def test_signatures () :
#     """ 
#     Test if signature violations are flagged 
#     """ 
# 
#     try                   : ret = sigtest ('string', 2.4, 'hallo') 
#     except TypeError as e : pass 
#     except Exception as e : assert (False), "TypeError != %s (%s)" % (type(e), e)
#     else                  : assert (False), "expected TypeError exception, got none"
# 
#     try                   : ret = sigtest ('string', 2, 1.1414) 
#     except Exception as e : assert (False), "exception %s: %s" % (type(e), e)
#
#
# # ------------------------------------------------------------------------------
# # run tests if called directly
# if __name__ == "__main__":
# 
#     test_signatures ()
# 
# # ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import re


# ------------------------------------------------------------------------------
#
class ReSult (object) :
    """
    This class is a container around a regular expression match, which provides
    some more conventient access methods, boolean tests, etc.

    We only handle base strings, not unicode strings!
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, result=None) :
        """
        construct with a `re.MatchObject` instance.  This ctor should only be
        called from within the `ReString` class.
        """
        self._glist = list()
        self._gdict = dict()


        if  result : 
            if  not isinstance (result, type(re.match("",""))) :  # fuck python 
                raise TypeError ('Need re.MatchObject on construction, not %s' % type(result))

            self._glist = result.groups ()
            self._gdict = result.groupdict ()


    # --------------------------------------------------------------------------
    #
    def __str__ (self) :
        """
        The string representation is based on the match *list*, as the dict may
        not include all matches...
        """
        return str(self._glist)


    # --------------------------------------------------------------------------
    #
    def __len__ (self) :
        """
        The len representation is based on the match *list*, as the dict may
        not include all matches...
        """
        return len(self._glist)


    # --------------------------------------------------------------------------
    #
    def get (self, key, default=None) :
        """
        get is supported for default based dict access,
        """
   
        if  isinstance (key, str) :
            return self._gdict.get (key, default)
        else :
            raise TypeError ("key %s needs to be integer, not %s"
                          % (key, type(key)))


    # --------------------------------------------------------------------------
    #
    def __getitem__ (self, idx) :
        """
        getitem is supported for both array type access (using an int index),
        and for dict type access (using a string name).  All other key types
        will cause an exception.
        """
   
        if  isinstance (idx, str) :
            if  idx in self._gdict :
                return self._gdict[idx]
        elif  isinstance (idx, int) :
            if  len(self) > idx :
                return self._glist[idx]
        else :
            raise TypeError ("index %s needs to be integer or string, not %s"
                          % (idx, type(idx)))
        return None


    # --------------------------------------------------------------------------
    # 
    def __iter__(self):
        """
        the matches can be iterated over
        """
        for m in self._glist :
            yield m

    # --------------------------------------------------------------------------
    #
    def __getattr__ (self, name) :
        """
        Matches can be accessed as properties
        """
        return self[name]


    # --------------------------------------------------------------------------
    #
    def __nonzero__ (self) :
        """
        Boolean check for 'if / elif / else' constructs 
        """
        if  len(self) :
            return True
        return False


    # --------------------------------------------------------------------------
    #
    def __enter__ (self) :
        """
        support context manager interface for with-statement based constructs
        """
        return self


    # --------------------------------------------------------------------------
    #
    def __exit__ (self, a, b, c) :
        """
        second part of the context manager interface
        """
        pass



    # --------------------------------------------------------------------------
    #
    def __cmp__ (self, other) :
        """
        compare to another ReSult or to a tuple.  As they are both iterable, we
        compare based on the iterable interface
        """
        if  isinstance (other, str) :
            # we consider a single string to be an interable of one element
            other = [str(other)]

        import collections 
        if  isinstance (other, collections.Iterable) :

            if  len (self) != len (other) :
                return len (self) - len (other)

            for i, m in enumerate (self) :
                if m != other[i] :
                  # print '%s != %s' % (m, other[i])
                    return len(m) - len(other[i])
            return 0

        raise TypeError ('compare expected iterable or single string type, not %s' % type (other))



# ------------------------------------------------------------------------------
class ReString (str) :
    """
    This is a string class which supports simplified regular expression
    matching. It is not thought that the regex language or expressions are
    simplified, but rather that the invokation of the matching is simple, as is
    the handling of the match results:

        txt = ReString ("The quick brown fox jumps over the lazy dog")
        
        # the '//' operator is overloaded to match against a regular expression.
        # The result is a `ReSult` instance, which allows simple access to the
        # matches
        with txt // '(\s.u)(?P<x>.*?j\S+)' as res :
            if res : print 'Matched!'               # boolean check
            print "res      : '%s' " % res          # list of results
            print "res[0]   : '%s' " % res[0]       # index by number ...
            print "res[1]   : '%s' " % res[1]       # ... for all matches
            print "res['x'] : '%s' " % res['x']     # index by match name
            print "res.x    : '%s' " % res.x        # ...   as properties
            for i, r in enumerate (res) :
                print "res %d    : '%s' " % (i, r)  # matches as iterable
        
            assert (len(res) == 2)                  # number of matches
            assert (res == [' qu', 'ick brown fox jumps'])  # compare to list
        
        if  txt // '(rabbit)' :                     # simple use in if / elif / ...
            res = txt.get ()                        # get ReSult of last match
        
        elif  txt // '((?:\s).{12,15}?(\S+))' :     # support for full Python regex slang
            res = txt.get ()
        
        else :
            print 'no match'
    """

    # --------------------------------------------------------------------------
    def __init__ (self, src) :

        self._result = None

        str.__init__ (self, src)


    # --------------------------------------------------------------------------
    def __floordiv__ (self, regex) :

        compiled_regex = None
        if  isinstance (regex, basestring) :
            compiled_regex = re.compile (regex)
        else :
            # assume we got a compiled regex
            # FIXME: type check
            compiled_regex = regex

        if re :
            self._result = ReSult (re.search (compiled_regex, self))
            return self._result

        return None


    # --------------------------------------------------------------------------
    def get (self, key=None, default=None) :

        if  self._result and key :

            try :
                return self._result[key]

            except KeyError :
                if  default :
                    return default
                raise

        return self._result


# ------------------------------------------------------------------------------
# 
def _example_re_string () :

    txt = ReString ("The quick brown fox jumps over the lazy dog")
    
    with txt // '(\s.u)(?P<x>.*?j\S+)' as res :

        if res : print 'Matched!'               # boolean check
        print "res      : '%s' " % res          # list of results
        print "res[0]   : '%s' " % res[0]       # index by number ...
        print "res[1]   : '%s' " % res[1]       # ... for all matches
        print "res['x'] : '%s' " % res['x']     # index by match name
        print "res.x    : '%s' " % res.x        # ...   as properties
        for i, r in enumerate (res) :
            print "res %d    : '%s' " % (i, r)  # matches as iterable

        assert (len(res) == 2)                  # number of matches
        assert (res == [' qu', 'ick brown fox jumps'])  # compare to list
    
    
    if  txt // '(rabbit)' :                     # simple use in if / elif / ...
        res = txt.get ()                        # get ReSult of last match
    
    elif  txt // '((?:\s).{12,15}?(\S+))' :     # support for full Python regex slang
        res = txt.get ()
    
    else :
        print 'no match'


# ------------------------------------------------------------------------------
# 
def _test_re_string () :

    txt   = ReString ("The quick brown fox jumps over the lazy dog")
    tgt_l = [' qu', 'ick brown fox jumps']
    tgt_d = {'x'  : 'ick brown fox jumps'}
    
    with txt // '(\s.u)(?P<x>.*?j\S+)' as res :
        assert (res)
        assert (len(res) == len(tgt_l))
        assert (res      == tgt_l), "%s != %s" % (str(res), str(tgt_l))
        assert (res[0]   == tgt_l[0])
        assert (res[1]   == tgt_l[1])
        assert (res['x'] == tgt_d['x'])
        assert (res.x    == tgt_d['x'])
        for i, r in enumerate (res) :
            assert (r    == tgt_l[i])
    
    if  txt // '(rabbit)' :
        assert (False)
    
    elif  txt // '((?:\s).{12,15}?(\S+))' :     # support for full Python regex slang
        assert (True)
    
    else :
        assert (False)


# ------------------------------------------------------------------------------

# _example_re_string()
# _test_re_string()

# ------------------------------------------------------------------------------




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>

"""
This module implements a pthread_atfork() work-a-like mechanism for all
fork() calls made from the Python os module.  Any time a fork() is called
from Python a set of unique callbacks will be made in each of the following
three states:
    Preparing to fork - Immediately before the fork call is made.
    In the parent after fork - Immediately after the fork (regardless of
                               success or failure) in the parent process.
    In the child after fork - Immediately after the fork in the child process.

To use this module, first import it early on your programs initialization:

    import atfork
    atfork.monkeypatch_os_fork_functions()

That will stub out os.fork and os.forkpty with wrapped versions implementing
the enhanced behavior.

Next, register your atfork actions by calling atfork.atfork:

    atfork.atfork(prepare=my_lock.acquire,
                  parent=my_lock.release,
                  child=my_lock.release)

No API to unregister an atfork call is provided.  If you are concerned
about resource usage by references your callable holds, consider using
weakref's within your callable.
"""

from .atfork       import monkeypatch_os_fork_functions, atfork
from .stdlib_fixer import fix_logging_module

fix_logging_module()
monkeypatch_os_fork_functions()






import os
import sys
import time
import pprint
import signal
import thread
import random
import threading
import traceback

from .threads import gettid

_ps_cmd = 'ps -efw'
if sys.platform != 'darwin':
    _ps_cmd += ' --forest'

# --------------------------------------------------------------------
#
def get_trace():

    trace = sys.exc_info ()[2]

    if  trace :
        stack           = traceback.extract_tb  (trace)
        traceback_list  = traceback.format_list (stack)
        return "".join (traceback_list)

    else :
        stack           = traceback.extract_stack ()
        traceback_list  = traceback.format_list (stack)
        return "".join (traceback_list[:-1])


# ------------------------------------------------------------------------------
#
class DebugHelper (object) :
    """
    When instantiated, and when "RADICAL_DEBUG" is set in the environment, this
    class will install a signal handler for SIGUSR1.  When that signal is
    received, a stacktrace for all threads is printed to stdout.
    We also check if SIGINFO is available, which is generally bound to CTRL-T.

    Additionally, a call to 'dh.fs_block(info=None)' will create a file system
    based barrier: it will create a unique file in /tmp/ (based on 'name' if
    given), and dump the stack trace and any 'info' into it.  It then waits
    until that file has changed (touched or removed etc), and then returns.  The
    wait is a simple pull based 'os.stat()' (once per sec).
    """

    # --------------------------------------------------------------------------
    #
    def __init__ (self, name=None, info=None):
        """
        name: string to identify fs barriers
        info: static info to dump into fs barriers
        """

        self._name       = name
        self._info       = info

        if not self._name:
            self._name = str(id(self))

        if 'MainThread' not in threading.current_thread().name:
            # python only supports signals in main threads :/
            return

        if 'RADICAL_DEBUG' in os.environ :
            signal.signal(signal.SIGUSR1, print_stacktraces) # signum 30
            signal.signal(signal.SIGQUIT, print_stacktraces) # signum  3

            try:
                assert signal.SIGINFO
                signal.signal(signal.SIGINFO, print_stacktraces) # signum 29
            except AttributeError as e:
                pass


    # --------------------------------------------------------------------------
    #
    def fs_block(self, info=None):
        """
        Dump state, info in barrier file, and wait for it tou be touched or
        read or removed, then continue.  Leave no trace.
        """

        if not 'RADICAL_DEBUG' in os.environ:
            return

        try:
            pid = os.getpid()
            tid = threading.currentThread().ident

            fs_barrier = "/tmp/ru.dh.%s.%s.%s" % (self._name, pid, tid)
            fs_handle  = open(fs_barrier, 'w+')
            fs_handle.seek(0,0)
            fs_handle.write("\nSTACK TRACE:\n%s\n%s\n" % (time.time(), get_trace()))
            fs_handle.write("\nSTATIC INFO:\n%s\n\n" % pprint.pformat(self._info))
            fs_handle.write("\nINFO:\n%s\n\n" % pprint.pformat(info))
            fs_handle.flush()

            new = os.stat(fs_barrier)
            old = new

            while old == new:
                new = os.stat(fs_barrier)
                time.sleep(1)

            fs_handle.close()
            os.unlink(fs_barrier)

        except Exception as e:
            # we don't care (much)...
            pass


# ------------------------------------------------------------------------------
#
def print_stacktraces(signum=None, sigframe=None):
    """
    signum, sigframe exist to satisfy signal handler signature requirements
    """

    this_tid = threading.currentThread().ident

    # if multiple processes (ie. a process group) get the signal, then all
    # traces are mixed together.  Thus we waid 'pid%100' milliseconds, in
    # the hope that this will stagger the prints.
    pid = int(os.getpid())
    time.sleep((pid%100)/1000)

    out  = "===============================================================\n"
    out += "RADICAL Utils -- Debug Helper -- Stacktraces\n"
    out += os.popen('%s | grep " %s " | grep -v grep' % (_ps_cmd, pid)).read()
    try:
        info = get_stacktraces()
    except Exception as e:
        out += 'skipping frame (%s)' % e
        info = None

    if info:
        for tid,tname in info:

            if tid == this_tid : marker = '[active]'
            else               : marker = ''
            out += "---------------------------------------------------------------\n"
            out += "Thread: %s %s\n" % (tname, marker)
            out += "  PID : %s \n"   % os.getpid()
            out += "  TID : %s \n"   % tid
            for fname,lineno,method,code in info[tid,tname]:

                if code: code = code.strip()
                else   : code = '<no code>'

              # # [:-1]: .py vs. .pyc :/
              # if not (__file__[:-1] in fname and \
              #         method in ['get_stacktraces', 'print_stacktraces']):
                if method not in ['get_stacktraces', 'print_stacktraces']:
                    out += "  File: %s, line %d, in %s\n" % (fname, lineno, method)
                    out += "        %s\n" % code

    out += "==============================================================="

    sys.stdout.write("%s\n" % out)

    if 'RADICAL_DEBUG' in os.environ:
        with open('/tmp/ru.stacktrace.%s.log' % pid, 'w') as f:
            f.write ("%s\n" % out)


# --------------------------------------------------------------------------
#
def get_stacktraces():

    id2name = {}
    for th in threading.enumerate():
        id2name[th.ident] = th.name

    ret = dict()
    for tid, stack in sys._current_frames().items():
        if tid in id2name:
            ret[tid,id2name[tid]] = traceback.extract_stack(stack)
        else:
            ret[tid,'noname'] = traceback.extract_stack(stack)

    return ret


# --------------------------------------------------------------------------
#
def print_stacktrace(msg=None):

    if not msg:
        msg = ''

    pid  = int(os.getpid())
    out  = "--------------\n"
    out += "RADICAL Utils -- Stacktrace [%s] [%s]\n" % (pid, threading.currentThread().name)
    out += "%s\n" % msg
    out += os.popen('%s | grep " %s " | grep -v grep' % (_ps_cmd, pid)).read()
    for line in get_stacktrace():
        out += line.strip()
        out += "\n"
    out += "--------------\n"

    sys.stdout.write(out)


# --------------------------------------------------------------------------
#
def print_exception_trace(msg=None):

    if not msg:
        msg = ''

    pid  = int(os.getpid())
    out  = "--------------\n"
    out += "RADICAL Utils -- Stacktrace [%s] [%s]\n" % (pid, threading.currentThread().name)
    out += "%s\n" % msg
    out += os.popen('%s | grep " %s " | grep -v grep' % (_ps_cmd, pid)).read()
    for line in traceback.format_exc().split('\n'):
        out += line.strip()
        out += "\n"
    out += "--------------\n"

    sys.stdout.write(out)


# --------------------------------------------------------------------------
#
def get_stacktrace():

    return traceback.format_stack()[:-1]


# ------------------------------------------------------------------------------
#
def get_caller_name(skip=2):
    """
    Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height

    Kudos: http://stackoverflow.com/questions/2654113/ \
            python-how-to-get-the-callers-method-name-in-the-called-method
    """
    import inspect

    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ''

    parentframe = stack[start][0]

    name   = []
    module = inspect.getmodule(parentframe)

    # `modname` can be None when frame is executed directly in console
    # TODO(techtonik): consider using __main__
    if module:
        name.append(module.__name__)

    # detect classname
    if 'self' in parentframe.f_locals:
        name.append(parentframe.f_locals['self'].__class__.__name__)

    codename = parentframe.f_code.co_name

    if codename != '<module>':  # top level usually
        name.append(codename)   # function or a method

    del parentframe

    return ".".join(name)


# ------------------------------------------------------------------------------
#
_verb  = False
if 'RADICAL_DEBUG' in os.environ :
    _verb = True

_raise_on_state = dict()
_raise_on_lock  = threading.Lock()

def raise_on(tag, log=None, msg=None):
    """
    The purpose of this method is to artificially trigger error conditions for
    testing purposes, for example when handling the n'th unit, getting the n'th
    heartbeat signal, etc.  

    The tag parameter is interpreted as follows: on the `n`'th invocation of
    this method with any given `tag`, an exception is raised, and the counter
    for that tag is reset.
    
    The limit `n` is set via an environment variable `RU_RAISE_ON_<tag>`, with
    `tag` in upper casing.  The environment will only be inspected during the
    first invocation of the method with any given tag.  The tag counter is
    process-local, but is shared amongst threads of that process.
    """

    global _raise_on_state
    global _raise_on_lock

    with _raise_on_lock:

        if tag not in _raise_on_state:
            env = os.environ.get('RU_RAISE_ON_%s' % tag.upper())
            if env and env.startswith('RANDOM_'):
                # env is rnd spec
                rate  = int(env[7:])
                limit = 1
            elif env:
                # env is int
                rate  = 1
                limit = int(env)
            else:
                # no env set
                rate  = 1
                limit = 0

            _raise_on_state[tag] = { 
                    'count' : 0,
                    'rate'  : rate,
                    'limit' : limit
                    }
            
        _raise_on_state[tag]['count'] += 1

        count = _raise_on_state[tag]['count']
        limit = _raise_on_state[tag]['limit']
        rate  = _raise_on_state[tag]['rate']

        if msg    : info = '%s [%2d / %2d] [%s]' % (tag, count, limit, msg)
        elif _verb: info = '%s [%2d / %2d]'      % (tag, count, limit     )

        if log    : log.debug('raise_on checked   %s' , info)
        elif _verb: print     'raise_on checked   %s' % info

        if limit and count == limit:
            _raise_on_state[tag]['count'] = 0
            if rate == 1:
                val = limit
            else:
                val = random.randint(0, 100)
                if val > rate:
                    if log:     log.warn('raise_on ignored   %s [%2d / %2d]' % (tag, val, rate))
                    elif _verb: print   ('raise_on ignored   %s [%2d / %2d]' % (tag, val, rate))
                    return

            if log:     log.warn('raise_on triggered %s [%2d / %2d]' % (tag, val, rate))
            elif _verb: print    'raise_on triggered %s [%2d / %2d]' % (tag, val, rate)

            # reset counter and raise exception
            raise RuntimeError('raise_on for %s [%s]' % (tag, val))


# ------------------------------------------------------------------------------
#
def attach_pudb(logger=None):

    host = '127.0.0.1'
  # host = gethostip()
    tid  = gettid()
    port = tid + 10000

    if logger:
        logger.info('debugger open: telnet %s %d', host, port)
    else:
        print 'debugger open: telnet %s %d' % (host, port)

    import pudb
    import signal
    pudb.DEFAULT_SIGNAL = signal.SIGALRM

    from pudb.remote import set_trace
    set_trace(host=host, port=port, term_size=(200, 50))


# ------------------------------------------------------------------------------




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>

"""
This module provides code to setup appropriate atfork() calls within the
Python standard library in order to make it safe for use in programs that
mix fork() and threads.

Import this and call the appropriate fixers very early in your program's
initialization sequence, preferrably -before- importing the modules you
need to fix.

Provided fixers:

    fix_logging_module()

TODO(gps): Audit more of the stdlib and provide necessary fixers.
In 2.4.5 the following additional stdlib modules use locks:

  threading.Condition and similar classes could use it.
  Queue
  cookielib
  mimetools
  _strptime
"""

import os
import sys
import warnings

from .atfork import atfork


class Error(Exception):
    pass


def fix_logging_module():

    import os
    if 'RADICAL_UTILS_NOATFORK' in os.environ:
        return

    logging = sys.modules.get('logging')
    # Prevent fixing multiple times as that would cause a deadlock.
    if logging and getattr(logging, 'fixed_for_atfork', None):
        return
    if logging:
        if os.environ.get('RADICAL_DEBUG'):
            warnings.warn('logging module already imported before fixup.')
    import logging
    if logging.getLogger().handlers:
        # We could register each lock with atfork for these handlers but if
        # these exist, other loggers or not yet added handlers could as well.
        # Its safer to insist that this fix is applied before logging has been
        # configured.
        warnings.warn('logging handlers already registered.')
      # raise Error('logging handlers already registered.')

    logging._acquireLock()
    try:
        def fork_safe_createLock(self):
            self._orig_createLock()
            atfork(self.lock.acquire,
                   self.lock.release, self.lock.release)

        # Fix the logging.Handler lock (a major source of deadlocks).
        logging.Handler._orig_createLock = logging.Handler.createLock
        logging.Handler.createLock = fork_safe_createLock

        # Fix the module level lock.
        atfork(logging._acquireLock,
               logging._releaseLock, logging._releaseLock)

        logging.fixed_for_atfork = True
    finally:
        logging._releaseLock()

# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils as ru


# ------------------------------------------------------------------------------
def test_plugin_manager () :
    """
    Test plugin manager
    """

    pmgr = ru.PluginManager ('radical.utils')

    plugin_2 = pmgr.load ('unittests_1', 'default_2')
    plugin_2.init ('a', 1)
    ret = plugin_2.run ()
    assert (ret == (1, 'a')), "plugin_2 invocation: '%s' != '%s'" % ([1, 'a'], ret)

    try :
        plugin_1 = pmgr.load ('unittests_1', 'default_1')
        assert False 
    except LookupError :
        pass
    except :
        assert False


    plugin_2 = pmgr.load ('unittests_1', 'default_2')
    plugin_2.init ('a', 1)
    ret = plugin_2.run ()
    assert (ret == (1, 'a')), "plugin_2 invocation: '%s' != '%s'" % ([1, 'a'], ret)
    
    plugin_3 = pmgr.load ('unittests_2', 'default_2')
    plugin_3.init ('a', 1)
    ret = plugin_3.run ()
    assert (ret == (1, 'a')), "plugin_3 invocation: '%s' != '%s'" % ([1, 'a'], ret)
    
    # load twice -- plugin_2 is marked as singleton plugin_2, and will raise if it
    # is created twice
    plugin_3 = pmgr.load ('unittests_2', 'default_2')
    plugin_3.init ('a', 1)
    ret = plugin_3.run ()
    assert (ret == (1, 'a')), "plugin_3 invocation: '%s' != '%s'" % ([1, 'a'], ret)
    

    try :
        pmgr = ru.PluginManager     ('troy')
        plugin_strategy = pmgr.load ('strategy', 'basic_late_binding')
        import troy
        session = troy.Session ()
        plugin_strategy.init_plugin (session)
        plugin_strategy.execute     (None, None, None, None)
    except :
        pass


    import resource
    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print "%5d  %d" % (0, mem)

    pmgr = ru.PluginManager ('radical.utils')

    import resource
    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print "%5d  %d" % (0, mem)

    for i in range (1000000) :
        plugin_4 = pmgr.load ('unittests_2', 'default_2')
        plugin_4.init ('a', 1)
        ret = plugin_4.run ()
        assert (ret == (1, 'a')), "plugin_4 invocation: '%s' != '%s'" % ([1, 'a'], ret)

        if not i % 100000 :
            mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            print "%5d  %d" % (i, mem)

    mem = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print "%5d  %d" % (i+1, mem)



# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_plugin_manager ()

# ------------------------------------------------------------------------------




#!/usr/bin/env python

import sys

import radical.utils as ru


# ------------------------------------------------------------------------------

def test_reporter():

    r = ru.Reporter(title='test')

    r.header  ('header  \n')
    r.info    ('info    \n')
    r.progress('progress\n')
    r.ok      ('ok      \n')
    r.warn    ('warn    \n')
    r.error   ('error   \n')
    r.plain   ('plain   \n')

    i = 0
    j = 0
    for cname,col in r.COLORS.items():
        if cname == 'reset':
            continue
        i += 1
        for mname,mod in r.COLOR_MODS.items():
            if mname == 'reset':
                continue
            j += 1
            sys.stdout.write("%s%s[%12s-%12s] " % (col, mod, cname, mname))
            sys.stdout.write("%s%s" % (r.COLORS['reset'], r.COLOR_MODS['reset']))
        sys.stdout.write("\n")
        j = 0

    import time

    r.info('test idler:')
    r.idle(mode='start')
    for i in range(3):
        r.idle()
        time.sleep(0.3)
    r.idle(color='ok', c='.')
    r.idle(color='error', c='.')
    for i in range(3):
        r.idle()
        time.sleep(0.1)

    r.idle(mode='stop')
    r.ok('>>done\n')

    r.info('idle test\n')
    r.info('1234567891         2         3         4         5         6         7         8\n\t')
    r.info('.0.........0.........0.........0.........0.........0.........0.........0')
    r.idle(mode='start')
    for i in range(200):
        r.idle(); time.sleep(0.01)
        r.idle(); time.sleep(0.01)
        r.idle(); time.sleep(0.01)
        r.idle(); time.sleep(0.01)
        r.idle(color='ok', c="+")
    r.idle(mode='stop')

    r.set_style('error', color='yellow', style='ELTTTTMELE', segment='X')
    r.error('error')
    r.exit('exit', 1)

# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    test_reporter()


# ------------------------------------------------------------------------------





__author__    = "Radical.Utils Development Team (Andre Merzky, Ole Weidner)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


""" 
And example on using the radical.utils.config tools. 

This example will read config options fomr $HOME/.examples.cfg::
    
    [config]
    casing   = upper
    excluded = sparks,pftools
    
    [sp3.cd]
    exe  = /usr/local/bin/sp3
    args = no,idea

That setting can be overwritting via $EXAMPLE_CONFIG_CASING.

"""

import radical.utils.config  as ruc


# ------------------------------------------------------------------------------
#
# a set of pre-defined options
#
_config_options = [
    { 
    'category'      : 'config',
    'name'          : 'casing', 
    'type'          : str, 
    'default'       : 'default',
    'valid_options' : ['default', 'lower', 'upper'],
    'documentation' : "This option determines the casing of example's output",
    'env_variable'  : 'EXAMPLE_CONFIG_CASING'
    },
    { 
    'category'      : 'config',
    'name'          : 'excluded', 
    'type'          : list, 
    'default'       : '',
    'valid_options' : [],
    'documentation' : "This option determines set of excluded components",
    'env_variable'  : ''
    }
]

_sp3_options = [
    { 
    'category'      : 'sp3.cd',
    'name'          : 'exe', 
    'type'          : str, 
    'default'       : '/usr/bin/sp3',
    'valid_options' : [],
    'documentation' : "This option determines set sp3 executable ",
    'env_variable'  : ''
    },
    { 
    'category'      : 'sp3.cd',
    'name'          : 'args', 
    'type'          : list, 
    'default'       : '',
    'valid_options' : [],
    'documentation' : "This option determines set sp3 arguments ",
    'env_variable'  : ''
    }
]


# ------------------------------------------------------------------------------
#
class FancyEcho (ruc.Configurable): 
    """ 
    This example will evaluate the given configuration, and 
    """

    #-----------------------------------------------------------------
    # 
    def __init__(self):
        
        # set the configuration options for this object
        ruc.Configurable.__init__ (self, 'examples')
        ruc.Configurable.config_options (self, 'config', _config_options)
        ruc.Configurable.config_options (self, 'sp3.cd', _sp3_options)

        # use the configuration
        self._cfg = self.get_config ('config')

        self._mode = self._cfg['casing'].get_value ()
        print "mode: %s" % self._mode

        self._excl = self._cfg['excluded'].get_value ()
        print "excl: %s" % type(self._excl)
        print "excl: %s" % self._excl


        if  not 'sp3' in self._excl :
            print 'running sp3'

        if  not 'sparks' in self._excl :
            print 'running sparks'

        if  not 'pftools' in self._excl :
            print 'running pftools'

        # use sp3 configuration
        self._sp3 = self.get_config ('sp3.cd')
        print self._sp3['exe'].get_value ()
        print self._sp3['args'].get_value ()


    #-----------------------------------------------------------------
    # 
    def echo (self, source) :

        target = ""

        if  self._mode == 'default' :
            target = source

        elif self._mode == 'lower' :
            target = source.lower()

        elif self._mode == 'upper' :
            target = source.upper()

        else :
            target = 'cannot handle mode %s' % self._mode

        return target


# ------------------------------------------------------------------------------
#

fc  = FancyEcho ()
src = ''

while 'quit' != src :
    src = raw_input ('> ')
    tgt = fc.echo (src)
    print '  ' + tgt


# ------------------------------------------------------------------------------
#







__author__    = "Radical.Utils Development Team (Andre Merzky, Matteo Turilli)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import sys
import string
import singleton

# import colorama as c


# ------------------------------------------------------------------------------
#
class Reporter(object):

    # we want reporter style to be consistent in the scope of an application
    __metaclass__ = singleton.Singleton

    # COLORS = {'white'       : c.Style.BRIGHT    + c.Fore.WHITE   ,
    #           'yellow'      : c.Style.BRIGHT    + c.Fore.YELLOW  ,
    #           'green'       : c.Style.BRIGHT    + c.Fore.GREEN   ,
    #           'blue'        : c.Style.BRIGHT    + c.Fore.BLUE    ,
    #           'cyan'        : c.Style.BRIGHT    + c.Fore.CYAN    ,
    #           'red'         : c.Style.BRIGHT    + c.Fore.RED     ,
    #           'magenta'     : c.Style.BRIGHT    + c.Fore.MAGENTA ,
    #           'black'       : c.Style.BRIGHT    + c.Fore.BLACK   ,
    #           'darkwhite'   : c.Style.DIM       + c.Fore.WHITE   ,
    #           'darkyellow'  : c.Style.DIM       + c.Fore.YELLOW  ,
    #           'darkgreen'   : c.Style.DIM       + c.Fore.GREEN   ,
    #           'darkblue'    : c.Style.DIM       + c.Fore.BLUE    ,
    #           'darkcyan'    : c.Style.DIM       + c.Fore.CYAN    ,
    #           'darkred'     : c.Style.DIM       + c.Fore.RED     ,
    #           'darkmagenta' : c.Style.DIM       + c.Fore.MAGENTA ,
    #           'darkblack'   : c.Style.DIM       + c.Fore.BLACK   ,
    #           'reset'       : c.Style.RESET_ALL + c.Fore.RESET
    #       }

    COLORS     = {'reset'        : '\033[39m',
                  'black'        : '\033[30m',
                  'red'          : '\033[31m',
                  'green'        : '\033[32m',
                  'yellow'       : '\033[33m',
                  'blue'         : '\033[34m',
                  'magenta'      : '\033[35m',
                  'cyan'         : '\033[36m',
                  'lightgray'    : '\033[37m',
                  'darkgray'     : '\033[90m',
                  'lightred'     : '\033[91m',
                  'lightgreen'   : '\033[92m',
                  'lightyellow'  : '\033[93m',
                  'lightblue'    : '\033[94m',
                  'lightmagenta' : '\033[95m',
                  'lightcyan'    : '\033[96m',
                  'white'        : '\033[97m'}

    COLOR_MODS = {'reset'        : '\033[0m',
                  'bold'         : '\033[1m',
                  'underline'    : '\033[4m',
                  'blink'        : '\033[5m',
                  'inverse'      : '\033[7m',
                  ''             : ''}


    # define terminal colors and other output options
    TITLE    = 'bold lightblue'
    HEADER   = 'bold lightyellow'
    INFO     = 'lightblue'
    IDLE     = 'lightwhite'
    PROGRESS = 'lightwhite'
    OK       = 'lightgreen'
    WARN     = 'lightyellow'
    ERROR    = 'lightred'

    EMPTY   = ''
    DOTTED  = '.'
    SINGLE  = '-'
    DOUBLE  = '='
    HASHED  = '#'

    COLOR       = os.environ.get('RP_REPORT_COLOR', 'True')
    ANIME       = os.environ.get('RP_REPORT_ANIME', 'True')
    LINE_LENGTH = 80

    if COLOR.lower() in ['0', 'false', 'off']: COLOR = False
    else                                     : COLOR = True

    if ANIME.lower() in ['0', 'false', 'off']: ANIME = False
    else                                     : ANIME = True


    # --------------------------------------------------------------------------
    #
    def __init__(self, title=None, targets=['stdout']):

        '''
        settings.style:
          E : empty line
          T : tabulator
          L : line of line segments
          M : text to report
        '''

        self._title    = title
        self._pos      = 0
        self._targets  = targets
        self._settings = {'title'    : {'color'   : self.TITLE,
                                        'style'   : 'ELMLE',
                                        'segment' : self.DOUBLE
                                       },
                          'header'   : {'color'   : self.HEADER,
                                        'style'   : 'ELME',
                                        'segment' : self.SINGLE
                                       },
                          'info'     : {'color'   : self.INFO,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'idle'     : {'color'   : self.IDLE,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'progress' : {'color'   : self.PROGRESS,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'ok'       : {'color'   : self.OK,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'warn'     : {'color'   : self.WARN,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'error'    : {'color'   : self.ERROR,
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       },
                          'plain'    : {'color'   : '',
                                        'style'   : 'M',
                                        'segment' : self.EMPTY
                                       }
                         }

        self._idle_sequence = '/-\\|'
        self._idle_pos      = dict()

        # set up the output target streams
        self._color_streams = list()
        self._streams       = list()

        for tgt in self._targets:

            if  tgt.lower() == 'stdout':
                self._color_streams.append(sys.stdout)

            elif tgt.lower() == 'stderr':
                self._color_streams.append(sys.stderr)

            else:
                # >>&     color stream in append    mode
                # >&      color stream in overwrite mode (default)
                # >>  non-color stream in append    mode
                # >   non-color stream in overwrite mode

                if   tgt.startswith('>>&'): self._color_streams.append(open(tgt[3:], 'a'))
                elif tgt.startswith('>&') : self._color_streams.append(open(tgt[2:], 'w'))
                elif tgt.startswith('>>') : self._streams.append      (open(tgt[2:], 'a'))
                elif tgt.startswith('>')  : self._streams.append      (open(tgt[1:], 'w'))
                else                      : self._color_streams.append(open(tgt,     'w'))

        # and send the title to all streams
        if self._title:
            self.title(self._title)


    # --------------------------------------------------------------------------
    #
    def _out(self, color, msg):

        if self.COLOR:
            color_mod = ''
            if  ' ' in color:
                color_mod, color = color.split(' ', 2)

            color     = self.COLORS.get(color.lower(), '')
            color_mod = self.COLOR_MODS.get(color_mod.lower(), '')

            color += color_mod


        # make sure we count tab length on line start correctly
        msg = msg.replace('\n\t', '\n        ')

        # make sure we don't extent a long line further
        if self._pos >= (self.LINE_LENGTH) and msg and msg[0] != '\n':
            while msg[0] == '\b':
                msg = msg[1:]
            msg = '\n        %s' % msg

        # special control characters:
        #
        #   * '>>' will, at it's place, insert sufficient spaces to make the
        #     remainder of the string right-aligned.  Only one >> is
        #     interpreted, linebreaks before it are ignored
        #
        #   * '<<' will insert a line break if the position is not already on
        #     the beginning of a line
        #
      # print "[%s:%s]" % (self.__hash__(), self._pos),
        slash_f  = msg.find('>>')
        if slash_f >= 0:
            copy   = msg[slash_f+1:].strip()
            spaces = self.LINE_LENGTH - self._pos - len(copy) + 1 # '>>'
            if spaces < 0:
                spaces = 0
            msg = msg.replace('>>', spaces * ' ')

        slash_cr = msg.find('<<')
        if slash_cr >= 0:
          # print "{%s:%s}" % (self.__hash__(), self._pos),
            if self._pos + slash_cr > 0:
                spaces = self.LINE_LENGTH - self._pos - 1
                msg = msg.replace('<<', '%s\\\n' % (spaces * ' '))
            else:
                msg = msg.replace('<<', '')

        mlen  = len(filter(lambda x: x in string.printable, msg))
        mlen -= msg.count('\b')

      # print "<%s>" % (self._pos),
        # find the last \n and then count how many chars we are writing after it
        slash_n = msg.rfind('\n')
        if slash_n >= 0:
          # print "(%s" % (self._pos),
            self._pos = mlen - slash_n - 1
          # print ": %s)" % (self._pos),
        else:
          # print "'%s'[%s" % (msg, self._pos),
            self._pos += mlen
          # print ": %s]" % (self._pos),


        for stream in self._color_streams:
            if self.COLOR:
                stream.write(color)
            stream.write(msg)
            if self.COLOR:
                stream.write(self.COLORS['reset'])
                stream.write(self.COLOR_MODS['reset'])
            try:
                stream.flush()
            except Exception as e:
                pass

        for stream in self._streams:
            stream.write(msg)
            try:
                stream.flush()
            except Exception as e:
                pass

    # --------------------------------------------------------------------------
    #
    def _format(self, msg, settings=None):

        if not msg:
            msg = ''

        if not settings:
            settings = {}

        color   = settings.get('color',   '')
        style   = settings.get('style',   'M')
        segment = settings.get('segment', '')

        for c in style:

            if  c == 'M':
                self._out(color, "%s" % msg)

            if  c == 'T':
                self._out(color, "\t")

            elif c == 'E':
                self._out(color, "\n")

            elif c == 'L':
                if segment:
                    self._out(color, "%s\n" % (self.LINE_LENGTH * segment))


    # --------------------------------------------------------------------------
    #
    def set_style(self, which, color=None, style=None, segment=None):

        if which not in self._settings:
            raise LookupError('reporter does not support style "%s"' % which)

        settings = self._settings[which]

        if color  : settings['color']   = color
        if style  : settings['style']   = style
        if segment: settings['segment'] = segment


    # --------------------------------------------------------------------------
    #
    def title(self, title=''):

        if not title:
            title = self._title

        if title:
            fmt   = " %%-%ds\n" % (self.LINE_LENGTH-1)
            title = fmt % title

        self._format(title, self._settings['title'])


    # --------------------------------------------------------------------------
    #
    def header(self, msg=''):

        if msg:
            fmt = "%%-%ds\n" % self.LINE_LENGTH
            msg = fmt % msg

        self._format(msg, self._settings['header'])


    # --------------------------------------------------------------------------
    #
    def info(self, msg=''):

        self._format(msg, self._settings['info'])


    # --------------------------------------------------------------------------
    #
    def idle(self, c=None, mode=None, color=None, idle_id=None):

        if not self.ANIME:
            return

        if not idle_id:
            idle_id = 'default'

        if color: col = self._settings[color]['color']
        else    : col = self._settings['idle']['color']

        idx = 0
        if   mode == 'start': self._out(col, 'O')
        elif mode == 'stop' : self._out(col, '\b ')
        else:
            if not c:
                idx  = self._idle_pos.get(idle_id, 0)
                c    = self._idle_sequence[idx % len(self._idle_sequence)]
                idx += 1
                self._out(col, '\b%s' % c)
            else:
                idx += 1
                self._out(col, '\b%s|' % c)

        self._idle_pos[idle_id] = idx


    # --------------------------------------------------------------------------
    #
    def progress(self, msg=''):

        if not msg:
            msg = '.'
        self._format(msg, self._settings['progress'])


    # --------------------------------------------------------------------------
    #
    def ok(self, msg=''):

        self._format(msg, self._settings['ok'])


    # --------------------------------------------------------------------------
    #
    def warn(self, msg=''):

        self._format(msg, self._settings['warn'])


    # --------------------------------------------------------------------------
    #
    def error(self, msg=''):

        self._format(msg, self._settings['error'])


    # --------------------------------------------------------------------------
    #
    def exit(self, msg='', exit_code=0):

        self.error(msg)
        sys.exit(exit_code)


    # --------------------------------------------------------------------------
    #
    def plain(self, msg=''):

        self._format(msg)


# ------------------------------------------------------------------------------

